"""
 * The MIT License
 *
 * Copyright (c) 2024 Patrick Hammer
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * """


def Prettyprint_Plan(actionlist):
    return [_prettyaction(x) for x in actionlist[1:]]


def Prettyprint_AllActions(actionlist):
    return [_prettyaction(x) for x in actionlist]


def Prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule):
    actions_values_preconditions = rule[0]
    action = _prettyaction(actions_values_preconditions[0])
    precons = actions_values_preconditions[2:]
    print("<(", end="")
    print("vars=" + str(list(actions_values_preconditions[1])), end=", ")
    for i, x in enumerate(precons):
        print(f"{_prettyTriplet(x)}", end="")
        if i != len(precons) - 1:
            print(f", ", end="")
    scoreInc = f"score+={rule[1][3][0]}"
    keys = f"vars={list(rule[1][3][1:])}"
    if action == "^up":  # print actions with same width
        action += "   "
    if action == "^down" or action == "^left":
        action += " "
    print(",", action + str(")"), "=/> (" + keys + ", " + _prettyTriplet(rule[1]) + ", " + scoreInc + ")>.",
          Hypothesis_TruthValue(rule_evidence[rule]))


def _prettyTriplet(triplet):
    (y, x, value) = triplet[:3]
    if y >= 0: y = " " + str(y)
    if x >= 0: x = " " + str(x)
    return "shape[" + str(x) + "," + str(y) + f" ]='{value}'"


def _prettyaction(action):
    if str(action).find("<function ") > -1:
        return "^" + str(action).split("<function ")[1].split(" at")[0]
    else:
        return "^" + str(action)



"""
 * The MIT License
 *
 * Copyright (c) 2024 Patrick Hammer
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * """

import random
import sys
from copy import deepcopy

from numba import jit

from nace.prettyprint import Prettyprint_rule


def Hypothesis_UseMovementOpAssumptions(
        leftOp, rightOp, upOp, downOp, dropOp, DisableOpSymmetryAssumptionFlag
):
    """

    Register operations in case Euclidean space operation alignment assumptions should be exploited which helps data
    efficiency

    @param leftOp:
    @param rightOp:
    @param upOp:
    @param downOp:
    @param dropOp:
    @param DisableOpSymmetryAssumptionFlag:
    @return:
    """
    global left, right, up, down, drop, DisableOpSymmetryAssumption
    left, right, up, down, drop, DisableOpSymmetryAssumption = (
        leftOp,
        rightOp,
        upOp,
        downOp,
        dropOp,
        DisableOpSymmetryAssumptionFlag,
    )


@jit(nopython=True)
def Hypothesis_TruthValue(wpn):
    """
    # The truth value of a hypothesis can be obtained directly from the positive and negative evidence counter

    @param wpn:
    @return:
    """
    (wp, wn) = wpn
    frequency = wp / (wp + wn)
    confidence = (wp + wn) / (wp + wn + 1)
    return frequency, confidence


@jit(nopython=True)
def Hypothesis_TruthExpectation(tv):
    """
    # The truth expectation calculation based on the truth value (frequency, confidence) tuple

    @param tv: truth value made up from frequency and confidence
    @return:
    """
    (f, c) = tv
    return c * (f - 0.5) + 0.5


def Hypothesis_Choice(rule_evidence, rule1, rule2):
    """
    # When two hypotheses predict a different outcome for the same conditions, the higher truth exp one is chosen

    @param rule_evidence:
    @param rule1:
    @param rule2:
    @return:
    """
    if rule1 in rule_evidence and rule2 not in rule_evidence:  # dv added (no longer triggered)
        return rule1
    if rule1 not in rule_evidence and rule2 in rule_evidence:  # dv added (no longer triggered)
        return rule2
    if rule1 not in rule_evidence and rule2 not in rule_evidence:  # dv added (no longer triggered)
        # panic
        print("ERROR this suggests a logic error in calling code.")
        return rule2

    t1 = Hypothesis_TruthValue(rule_evidence[rule1])
    t2 = Hypothesis_TruthValue(rule_evidence[rule2])
    if Hypothesis_TruthExpectation(t1) > Hypothesis_TruthExpectation(t2):
        return rule1
    return rule2


def Hypothesis_Contradicted(rule_evidence, ruleset, negruleset, rule):  # this mutates returned copy of rule_evidence
    """
    # Negative evidence was found for the hypothesis/rule

    @param rule_evidence:
    @param ruleset:
    @param negruleset:
    @param rule:
    @return:
    """
    rule_evidence = _AddEvidence(rule_evidence, rule, False)  # mutates returned copy of rule_evidence
    if "silent" not in sys.argv:
        print("Neg. revised: ", end="")
        Prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule)
        # in a deterministic setting this would have sufficed however
        # simply excluding rules does not work in non-deterministic ones
        # if rule in ruleset:
        #    print("RULE REMOVAL: ", end=""); Prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule)
        #    ruleset.remove(rule)
        # negruleset.add(rule)
    return rule_evidence, ruleset, negruleset


def Hypothesis_Confirmed(  # this mutates the returned rule_evidence and ruleset
        FocusSet, rule_evidence, ruleset, negruleset, rule
):
    """
    # Positive evidence was found for the hypothesis/rule
    # Confirm rule against +ve and -ve evidence, add variants (i.e. transforms) to newrules

    @param FocusSet:
    @param rule_evidence: dict[rule:(+ve evidence, -ve evidence)]
    @param ruleset:
    @param negruleset:
    @param rule:
    @return:
    """

    rule_evidence = deepcopy(rule_evidence)  # dv added this line
    ruleset = deepcopy(ruleset)  # dv added this line
    # try location symmetry
    variants = _Variants(FocusSet, rule)
    for i, r in enumerate(variants):
        if i > 0:  # abduced hypotheses
            if r in rule_evidence:  # this derived hypothesis already exists
                continue
        rule_evidence = _AddEvidence(rule_evidence, r, True)  # mutates returned copy of rule_evidence
        if "silent" not in sys.argv:
            print("Pos. revised: ", end="")
            Prettyprint_rule(rule_evidence, Hypothesis_TruthValue, r)
        if r not in negruleset:
            if r not in ruleset:
                # print("RULE ADDITION: ", end=""); Prettyprint_rule(rule)
                ruleset.add(r)
    return rule_evidence, ruleset


def Hypothesis_ValidCondition(cond):
    """
    # Valid condition predicate defining the accepted neighbourhood between conclusion and condition cells
    # restrict to neighbours (CA assumption)
    # If 0,1, or 2 in distance in any direction, return True, else False

    @param cond:
    @return:
    """
    (y, x, v) = cond
    if y == 0 and x == 0:  # self
        return True
    if y == 0 and (x == -1 or x == -2):  # left
        return True
    if (y == -1 or y == -2) and x == 0:  # up
        return True
    if y == 0 and (x == 1 or x == 2):  # right
        return True
    if (y == 1 or y == 2) and x == 0:  # down
        return True
    return False


def Hypothesis_BestSelection(rules, rulesExcluded, rule_evidence,
                             stayed_the_same):  # mutates returned rules, rulesExcluded
    """

    We exclude rules which have more negative evidence than positive, and choose the highest truth-exp ones whenever
    a different outcome would be predicted for the same conditions

    @param rules:
    @param rulesExcluded:
    @param rule_evidence:
    @param stayed_the_same: World stayed the same when last ground truth copied in
    @return:
    """
    rulesin = deepcopy(rules)
    for i, rule1 in enumerate(rulesin):
        # if Hypothesis_TruthExpectation(Hypothesis_TruthValue(rule_evidence[rule1])) <= 0.5: #exclude rules which
        # are not better than exp (only 0.5+ makes sense here)
        if rule1 in rule_evidence:  # dv 8/jul/24 added to stop key not found which hadn't happened before
            if Hypothesis_TruthExpectation(
                    Hypothesis_TruthValue(rule_evidence[rule1])
            ) <= 0.5 or (
                    stayed_the_same
                    and random.random()
                    > Hypothesis_TruthExpectation(Hypothesis_TruthValue(rule_evidence[rule1]))
            ):
                if rule1 in rules:
                    rulesExcluded.add(rule1)
                    rules.remove(rule1)
    rulesin = deepcopy(rules)
    for i, rule1 in enumerate(rulesin):
        for j, rule2 in enumerate(rulesin):
            if (
                    i != j
            ):  # exclude rules of same precondition which are worse by truth value
                if rule1[0] == rule2[0]:
                    rulex = Hypothesis_Choice(rule_evidence, rule1, rule2)
                    if rulex == rule1:
                        if rule2 in rules:
                            rulesExcluded.add(rule2)
                            rules.remove(rule2)
                            # print("excluded ", end=''); Prettyprint_rule(rule2)
                    else:
                        if rule1 in rules:
                            rulesExcluded.add(rule1)
                            rules.remove(rule1)
                            # print("excluded", end=''); Prettyprint_rule(rule1)
    return rules, rulesExcluded


def _OpRotate(op):
    """
    # Rotate the operation in Euclidean space if Euclidean op assumptions are allowed to be used

    @param op:
    @return:
    """
    if op == right:
        return down
    if op == down:
        return left
    if op == left:
        return up
    if op == up:
        return right


def _ConditionRotate(cond):
    """
    # Rotate the conditions as well if Euclidean op assumptions are allowed to be utilized

    @param cond:
    @return:
    """
    (y, x, v) = cond
    if y == 0 and x == -1:  # left
        return (-1, 0, v)  # up
    if y == 0 and x == -2:  # left
        return (-2, 0, v)  # up
    if y == -1 and x == 0:  # up
        return (0, 1, v)  # right
    if y == -2 and x == 0:  # up
        return (0, 2, v)  # right
    if y == 0 and x == 1:  # right
        return (1, 0, v)  # down
    if y == 0 and x == 2:  # right
        return (2, 0, v)  # down
    if y == 1 and x == 0:  # down
        return (0, -1, v)  # left
    if y == 2 and x == 0:  # down
        return (0, -2, v)  # left
    if x == 0 and y == 0:
        return (0, 0, v)


def _Variants(
        focus_set, rule
):
    """
    # The rule variants, including hypothetical abduced variations for different directions based on Euclidean space
    rotation and "operation-independence" hypotheses.
    # Exploits Euclidean space properties (knowledge about World_Movement operations for faster learning)

    """

    global left, right, up, down, drop  # DV added

    action_values_precons = rule[0]
    conditions = action_values_precons[2:]
    action = action_values_precons[0]
    max_focus = None
    max_focus_val = False
    if len(focus_set) > 0:
        max_focus = max(focus_set, key=lambda k: focus_set[k])
    if max_focus is not None:
        for x, y, val in action_values_precons[2:]:
            if val == max_focus or rule[1][2] == max_focus:
                max_focus_val = True
    for y, x, v in conditions:
        if (action == left or action == right) and y != 0:
            return []
        if (action == up or action == down or action == drop) and x != 0:
            return []
    rules = [rule]
    action2 = _OpRotate(action)
    action3 = _OpRotate(action2)
    action4 = _OpRotate(action3)
    if DisableOpSymmetryAssumption:
        return rules
    if not max_focus_val:
        rules.append(
            (tuple([left, action_values_precons[1]] + list(conditions)), rule[1])
        )
        rules.append(
            (tuple([right, action_values_precons[1]] + list(conditions)), rule[1])
        )
        rules.append(
            (tuple([up, action_values_precons[1]] + list(conditions)), rule[1])
        )
        rules.append(
            (tuple([down, action_values_precons[1]] + list(conditions)), rule[1])
        )
    if (
            action != left and action != right and action != down and action != up
    ):  # not such an op where symmetry would apply
        return rules
    conditionlist2 = sorted([_ConditionRotate(x) for x in conditions])
    conditionlist3 = sorted([_ConditionRotate(x) for x in conditionlist2])
    conditionlist4 = sorted([_ConditionRotate(x) for x in conditionlist3])
    if max_focus_val:
        rules.append(
            (tuple([action2, action_values_precons[1]] + conditionlist2), rule[1])
        )
        rules.append(
            (tuple([action3, action_values_precons[1]] + conditionlist3), rule[1])
        )
        rules.append(
            (tuple([action4, action_values_precons[1]] + conditionlist4), rule[1])
        )
    return rules


def _AddEvidence(rule_evidence, rule, positive, w_max=20):  # Mutates a copy of rule_evidence
    """
    Add positive or negative evidence for a rule, with a certain max. amount of evidence so that non-stationary
    environments can be handled too

    @param rule_evidence:
    @param rule:
    @param positive:
    @param w_max:
    @return:
    """
    rule_evidence = deepcopy(rule_evidence)
    if rule not in rule_evidence:
        rule_evidence[rule] = (0, 0)
    (wp, wn) = rule_evidence[rule]
    if positive:
        if wp + wn <= w_max:
            rule_evidence[rule] = (wp + 1, wn)
        else:
            rule_evidence[rule] = (wp, max(0, wn - 1))
    else:
        if wp + wn <= w_max:
            rule_evidence[rule] = (wp, wn + 1)
        else:
            rule_evidence[rule] = (max(0, wp - 1), wn)
    return rule_evidence



"""
 * The MIT License
 *
 * Copyright (c) 2024 Dwane van der Sluis
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * """
import copy
import random
import sys
import time
from typing import Type

from nace.agent_module import Agent
from nace.nace_v3 import nacev3_get_next_action, nacev3_predict_and_observe
from nace.world_module_numpy import NPWorld


def _get_expected_deltas(nace_action_name):
    expected_directions = {
        "^left": (-1, 0),
        "^right": (1, 0),
        "^up": (0, -1),
        "^down": (0, 1)
    }
    expected_column_delta, expected_row_delta = expected_directions.get(nace_action_name, (0, 0))
    return expected_row_delta, expected_column_delta


class StepperV4():
    def __init__(self, seed_value=1, unobserved_code='.'):
        # The world is held externally.
        # The state of the agent (score number of keys held etc.) is held externally
        # Actions are performed on that world externally
        # this class wraps code and expectations that allow it to take one of the passed in worlds and predict forward.
        # partial observation (if it occurs) happens outside this code.

        # This module hole

        # variables set to specific values and passed in
        self.current_behavior = "BABBLE"
        random.seed(seed_value)

        # variables set to empty sets, or lists before the first call of this routine

        # these 4 var are global in the original implementation
        self.focus_set = dict([])
        self.rule_evidence = dict([])
        self.rules = set()
        self.negrules = set()
        self.plan = []
        self.time_counter = -1
        self.unobserved_code = unobserved_code  # indicates we have not seen the maps true value

        self.anamestr = "1234"

        # variables NOT set before the first call of this routine (in original implementation)
        self.used_rules = None
        self.debuginput = None

        self.rules_excluded = None
        self.stayed_the_same = False  # when last time the new ground truth is copied over it.

        self.post_action_agent = None
        self.pre_action_agent = None

        self.lastplanworld = None  # outputed
        self.planworld = None  # outputed
        self.internal_world = None  # working copy

        self.internal_preaction_world = None  # updated each time a best action is predicted
        self.action = None  # predicted best action

    def set_agent_ground_truth_state(self, xy_loc, score, values_exc_score):

        if self.post_action_agent is None:
            self.pre_action_agent = Agent(xy_loc, score, [0 for _ in values_exc_score])
            self.post_action_agent = Agent(xy_loc, score,
                                           values_exc_score=values_exc_score)
        else:
            self.pre_action_agent = copy.deepcopy(self.post_action_agent)
            self.post_action_agent.set_xy_loc(xy_loc)
            self.post_action_agent.set_values_inc_score([score] + list(values_exc_score))
        # Possible feature: try and reverse engineer rules that change each value into the description for that value
        # 'kx',left -> 'x ', v[1]+=1   --> pick up k
        result = {}
        for i, v in enumerate(self.post_action_agent.values):
            if i == 0:
                result["score"] = {"v": v}
            else:
                result["v" + str(i)] = {"v": v}
        return result

    def set_world_ground_truth_state(self, ground_truth_external_world: Type[NPWorld], new_xy_loc, time_counter):
        # Step 1: Update agent's field of view
        # world_module.World_FieldOfView(time_counter, xy_loc, internal_world_model, external_ground_truth_world_model)
        if ground_truth_external_world is not None:
            modified_count, _ = self.internal_world.update_world_from_ground_truth(
                time_counter,
                ground_truth_external_world, xy_locations=[new_xy_loc])

    def get_next_action(self, ground_truth_external_world: Type[NPWorld], new_xy_loc, print_debug_info,
                        time_delay_sec=0.0, available_actions=tuple([]), view_dist_x=3, view_dist_y=2):

        if self.pre_action_agent is None:
            self.pre_action_agent = Agent(new_xy_loc, score=0, values_exc_score=(0,))
        if self.internal_world is None:
            self.internal_world = NPWorld(with_observed_time=True, name="self.internal_world", view_dist_x=view_dist_x,
                                          view_dist_y=view_dist_y)

        start_time = time.time()
        self.time_counter += 1
        agent = self.post_action_agent if self.post_action_agent is not None else self.pre_action_agent

        self.plan, self.action, self.rules_excluded, self.current_behavior = nacev3_get_next_action(
            self.time_counter,
            self.focus_set,
            self.rule_evidence,  # can be mutated
            agent.get_xy_loc(),  #
            self.internal_world,  # passed in, has updates from external world applied to it.
            self.rules,
            ground_truth_external_world,  # used to update internal world
            print_debug_info=print_debug_info,
            stayed_the_same=self.stayed_the_same,
            agent=agent,
            available_actions=available_actions
        )

        # Store a copy after the external updates are copied in, i.e. the world use to predict on
        self.internal_preaction_world = copy.deepcopy(self.internal_world)

        end_time = time.time()
        if "manual" in sys.argv:
            print(self.current_behavior)
        else:
            if print_debug_info:
                print("get_next_action()", "focus=" + str(self.focus_set), self.current_behavior, "next_action=",
                      str(self.action))
        elapsed_time = end_time - start_time
        if (
                elapsed_time < time_delay_sec
                and "nosleep" not in sys.argv
                and "debug" not in sys.argv
                and "manual" not in sys.argv
        ):
            time.sleep(time_delay_sec - elapsed_time)

        return self.action, self.current_behavior

    def predict_and_observe(self,
                            object_count_threshold=1,
                            print_out_world_and_plan=True
                            ):
        """
        Possibly a new main entry point when the world does not need to be
        modeled in a way that actions can occur on it.
        @param object_count_threshold: how many new instances can be seen , and be considered, more are ignored
        @return:
        """
        (
            self.used_rules,
            self.focus_set,
            self.rule_evidence,
            predicted_world,  # self.t_minus_1_partially_observed_world,
            self.rules,
            self.negrules,
            values,
            self.lastplanworld,
            self.planworld,
            self.stayed_the_same
        ) = nacev3_predict_and_observe(
            self.time_counter,
            self.focus_set,
            self.rule_evidence,
            self.pre_action_agent.get_xy_loc(),
            copy.deepcopy(self.internal_preaction_world),  # pre action world (with times)
            self.rules,  # used in part1 and part2
            self.negrules,
            self.plan,
            self.action,
            self.rules_excluded,
            copy.deepcopy(self.internal_world),
            # post action world, can be internal or external, differences are copied in
            pre_action_agent=self.pre_action_agent,  # pre action agent
            ground_truth_post_action_agent=self.post_action_agent,  # post action agent
            unobserved_code=self.unobserved_code,
            object_count_threshold=object_count_threshold,
            print_out_world_and_plan=print_out_world_and_plan,
        )



import collections
import copy
import random
from collections import deque
from typing import Type

import nace.color_codes
import nace.world_module
from nace.agent_module import Agent
# from nace.hypothesis import *
from nace.hypothesis import Hypothesis_BestSelection, Hypothesis_ValidCondition, Hypothesis_Confirmed, Hypothesis_Contradicted
from nace.prettyprint import Prettyprint_AllActions
from nace.world_module_numpy import NPWorld


# See nace.py for type and data structure descriptions


def _plan(world: Type[NPWorld], rules, actions, focus_set,
          agent: Type[Agent], max_num_actions: int = 100,
          max_queue_len: int = 2000, custom_goal=None,
          continue_planning_threshold=0.99):
    """
    Plan forward searching for situations of highest reward or lowest AIRIS confidence or oldest age.

    Returns:
    - lowest_conf_actions: List of actions leading to the lowest AIRIS confidence
    - lowest_AIRIS_confidence: The lowest AIRIS confidence score (highest reward for exploring)
    - oldest_age_actions: Actions leading to the oldest world state
    - oldest_age: Age of the oldest world state encountered
    """
    queue = deque([(world, agent.get_values_inc_score(), [], 0, "na 1")])  # (world_state, actions, depth, debug_data)
    encountered = {}  # used to short-circuit search
    evaluation_count = 0

    lowest_conf_actions = []
    lowest_AIRIS_confidence = float("inf")  # smaller is better
    lowest_conf_predicted_score_delta = 0.0
    lowest_conf_stopping_reason = "na 2"

    oldest_age_actions = []
    oldest_age = 0  # if we see an age (time difference) greater than this, we store it.
    oldest_age_predicted_score_delta = 0.0
    oldest_age_stopping_reason = "na 3"

    dead_plans = []

    if agent.get_score() >= 1:
        pass

    while queue and len(queue) <= max_queue_len:  # queue size = number of actions * roll out length
        (current_world, current_agent_values, planned_actions, depth, stopping_reason) = queue.popleft()

        # need to store agent values on the queue as well, and add the values delta to them.

        if depth > max_num_actions:
            print(f"Max depth of {max_num_actions} reached, stopping search.")
            stopping_reason = "Max depth"
            dead_plans.append((planned_actions, stopping_reason))
            continue

        world_state = tuple([current_world.get_board_hashcode()] + current_agent_values[1:])

        if _should_skip_state(world_state, encountered, depth):
            # skip states already evaluated (optimisation)
            # print(f"Skipping state")
            stopping_reason = "Skipping state"
            dead_plans.append((planned_actions, stopping_reason))
            continue

        # store that we have been here, in order to prune the search space and avoid loops
        encountered[world_state] = depth

        for action in actions:  # check each action in turn from a known state
            new_world, new_AIRIS_confidence, new_age, agent_values_delta, predicted_score_delta = (
                _predict_next_world_state(
                    focus_set, current_world, action, rules, agent, custom_goal
                ))
            if agent_values_delta[0] != 0:
                pass  # code to place breakpoint on
            # if predicted_score_delta > 0.0:
            #     print("goal found") # code to place breakpoint on

            if (new_world.get_board_hashcode() == current_world.get_board_hashcode()):  # no effect (e.g. hit a wall )
                stopping_reason = "action triggered no change"
                dead_plans.append((planned_actions + [action], stopping_reason))
                continue  # no need to continue evaluating (optimisation)

            if (new_AIRIS_confidence == float("inf")):  # something bad happens, avoid.
                stopping_reason = "something bad happens (-ve score)"
                dead_plans.append((planned_actions + [action], stopping_reason))
                continue  # no need to continue evaluating (optimisation)

            new_planned_actions = planned_actions + [action]

            if _is_lower_airis_confidence(new_AIRIS_confidence, lowest_AIRIS_confidence, new_planned_actions,
                                          lowest_conf_actions):  # lower or equal confidence and fewer actions
                lowest_conf_actions = new_planned_actions
                lowest_AIRIS_confidence = new_AIRIS_confidence
                lowest_conf_predicted_score_delta = predicted_score_delta
                lowest_conf_stopping_reason = stopping_reason

            if _is_older_age(new_age, oldest_age, new_planned_actions, oldest_age_actions):
                oldest_age_actions = new_planned_actions
                oldest_age = new_age
                oldest_age_predicted_score_delta = predicted_score_delta
                oldest_age_stopping_reason = stopping_reason

            if new_AIRIS_confidence >= continue_planning_threshold:
                # calc value deltas together
                agent_values = list(copy.deepcopy(current_agent_values))
                for i, v in enumerate(agent_values_delta):
                    if i < len(agent_values):
                        agent_values[i] += v
                # add this to the queue to have extra steps added (confidence of 1 means certain, so keep planning)
                queue.append((new_world, agent_values, new_planned_actions, depth + 1, stopping_reason))
            else:
                # do not add this result to the queue, so it stops being explored further
                stopping_reason = "AIRIS confidence dropped below 1.0"
                dead_plans.append((planned_actions + [action], stopping_reason))

            evaluation_count += 1

            if new_AIRIS_confidence == float("-inf") and predicted_score_delta <= 0.0:
                print("logic error")
            if new_AIRIS_confidence > float("-inf") and predicted_score_delta > 0.0:
                print("logic error")

            if predicted_score_delta > 0.0:  # new_AIRIS_confidence == float("-inf"):
                # goal found?
                assert new_AIRIS_confidence == lowest_AIRIS_confidence
                assert predicted_score_delta > 0.0  # can we 'correct' the 'if' above to use score?
                assert lowest_conf_predicted_score_delta > 0.0 or oldest_age_predicted_score_delta > 0.0
                return (lowest_conf_actions,
                        lowest_AIRIS_confidence,
                        lowest_conf_predicted_score_delta,
                        lowest_conf_stopping_reason,
                        oldest_age_actions,
                        oldest_age,
                        oldest_age_predicted_score_delta,
                        oldest_age_stopping_reason)

    return (lowest_conf_actions, lowest_AIRIS_confidence, lowest_conf_predicted_score_delta, lowest_conf_stopping_reason,
            oldest_age_actions, oldest_age, oldest_age_predicted_score_delta, oldest_age_stopping_reason)


def _should_skip_state(world_state, encountered, depth):
    return world_state in encountered and depth >= encountered[world_state]


def _is_lower_airis_confidence(new_score, best_score, new_actions, best_actions):
    return (new_score < best_score or
            (new_score == best_score and len(new_actions) < len(best_actions)))


def _is_older_age(new_age, oldest_age, new_actions, best_actions):
    """
    Ages are the difference in the time counter. 1 is 1 time ago. -inf is the start of time.
    @param new_age:
    @param oldest_age:
    @param new_actions:
    @param best_actions:
    @return:
    """
    return (new_age > oldest_age or
            (new_age == oldest_age and len(new_actions) < len(best_actions)))


def _print_score(score):
    """
    Print score value taking its semantics regarding its value range semantics for planning into account

    @param score:
    @return:
    """
    if 0.0 <= score <= 1.0:
        print("certainty:", score)
    else:
        print("desired: True")


def nacev3_get_next_action(
        time_counter,
        focus_set,
        rule_evidence,
        xy_loc,
        internal_world_model: Type[NPWorld],
        rules_in,
        external_ground_truth_world_model: Type[NPWorld],
        print_debug_info,
        stayed_the_same: bool,
        agent: Type[Agent],
        available_actions: list  # used during babling
):
    """
    Determine the next action for the NACE agent based on current observations and rules.

    Steps:
    1. Limit the agent's field of view (partial observability)
    2. Refine rules based on evidence
    3. Plan forward to calculate favored actions and scores

    :param time_counter: Current time
    :param focus_set: Used in subroutine
    :param rule_evidence: Evidence for rules
    :param xy_loc: Tuple(x, y) of agent's location (0,0 is top left)
    :param internal_world_model: Agent's view of the world (mutated)
    :param rules_in: Input rules
    :param external_ground_truth_world_model: Current world state (mutated)
    :param  print_debug_info,
    :param stayed_the_same: bool,
    :param agent,
    :return: Tuple(plan, action, rules_excluded, behavior)


    Notes
        The number of state values that the agent stores, and the number of value delta in the rules must match.
    """
    rules_excluded = set()
    rules = copy.deepcopy(rules_in)

    # Step 1: Update agent's field of view
    # world_module.World_FieldOfView(time_counter, xy_loc, internal_world_model, external_ground_truth_world_model)
    if external_ground_truth_world_model is not None:
        modified_count, _ = internal_world_model.update_world_from_ground_truth(
            time_counter,
            external_ground_truth_world_model, xy_locations=[xy_loc])
        if print_debug_info:
            internal_world_model.multiworld_print([{"Caption": f"Internal:",
                                                    "World": internal_world_model,
                                                    "Color": nace.color_codes.color_code_white_on_black},
                                                   ]
                                                  )

    # Step 2: Refine rules based on evidence
    rules, rules_excluded = Hypothesis_BestSelection(
        rules,
        rules_excluded,
        rule_evidence,
        stayed_the_same)
    if agent.get_score() >= 1.0:
        pass

    # Step 3: Plan forward
    (lowest_conf_actions, lowest_AIRIS_confidence, lowest_conf_predicted_score_delta, lowest_conf_stopping_reason,
     oldest_age_actions, oldest_age, oldest_age_predicted_score_delta, oldest_age_stopping_reason) = (
        _plan(
            internal_world_model,
            rules,
            nace.world_module.get_full_action_list(),
            focus_set,
            agent=agent,
            custom_goal=nace.world_module.World_GetObjective()
        ))

    # Determine available actions
    remaining_actions = _get_remaining_actions(rules)
    all_actions_tried = len(remaining_actions) == 0

    # Determine behavior and action
    behavior, action, plan = _determine_behavior_and_action(
        lowest_AIRIS_confidence, lowest_conf_actions, lowest_conf_predicted_score_delta, lowest_conf_stopping_reason,
        oldest_age_actions, oldest_age, oldest_age_predicted_score_delta, oldest_age_stopping_reason,
        all_actions_tried,
        available_actions=available_actions,
        print_debug_info=print_debug_info
    )

    return plan, action, rules_excluded, behavior


def _get_remaining_actions(rules):
    action_list = nace.world_module.get_full_action_list()
    for rule in rules:
        precondition, _ = rule
        action = precondition[0]
        if action in action_list:
            action_list.remove(action)
    return action_list


# def _determine_behavior_and_action_old(airis_score, favored_actions, favored_actions_for_revisit, oldest_age,
#                                        all_actions_tried: bool, board_value_transition: tuple,
#                                        print_debug_info: bool = False):
#     """
#
#     Indeterministic if all_actions_tried == FALSE
#
#     @param airis_score:
#     @param favored_actions:
#     @param favored_actions_for_revisit:
#     @param oldest_age: oldest observed age (smaller == older)
#     @param all_actions_tried: True if all possible actions were in the set of applicable rules
#     @param babbling_rates: dict of rates e.g. : {'curiosity': 0.5, 'exploit': 1.0, 'explore': 0.5}
#     @param board_value_transition: tuple ( char, char) from to values of a board transition.
#     @param print_debug_info:
#     @return:
#     """
#
#     if all_actions_tried or airis_score == float("-inf"):
#         exploit_babble = False
#     else:
#         exploit_babble = random.random() > 0.5
#
#     if all_actions_tried:  # if all_actions_tried evaluates to False, else False 50% of time
#         explore_babble = False
#     else:
#         explore_babble = random.random() > 0.5
#
#     if airis_score >= 0.9 or exploit_babble or len(favored_actions) == 0:
#         # EXPLORE or BABBLE
#         if (not exploit_babble and not explore_babble and oldest_age > 0.0 and
#                 airis_score == 1.0 and
#                 len(favored_actions_for_revisit) != 0):
#             behavior = "EXPLORE"
#             print(behavior, Prettyprint_Plan(favored_actions_for_revisit), "oldest_age:", oldest_age)
#             action = favored_actions_for_revisit[0]
#             plan = favored_actions_for_revisit
#         else:
#             behavior = "BABBLE"
#             action = _choose_babble_action(available_actions=world_module.get_full_action_list())
#             plan = []
#     else:  # airis_score must be < 0.9
#         # ACHIEVE or CURIOUS
#         behavior = "ACHIEVE" if airis_score == float("-inf") else "CURIOUS"
#         print(behavior, Prettyprint_Plan(favored_actions), end=" ")
#         _print_score(airis_score)
#         action = favored_actions[0]
#         plan = favored_actions
#
#     if print_debug_info:
#         print("behavior", behavior,
#               "airis_score", airis_score,
#               "favored_actions", favored_actions,
#               "favored_actions_for_revisit", favored_actions_for_revisit, "oldest_age>0", (oldest_age > 0.0),
#               "board_value_transition", board_value_transition)
#
#     return behavior, action, plan
#

def _determine_behavior_and_action(lowest_AIRIS_conf,
                                   lowest_AIRIS_conf_actions,
                                   lowest_conf_predicted_score_delta:float,
                                   lowest_AIRIS_conf_stopping_reason: str,
                                   oldest_actions,
                                   oldest_age,
                                   oldest_age_predicted_score_delta:float,
                                   oldest_age_stopping_reason: str,
                                   all_actions_tried: bool,
                                   available_actions: list,
                                   print_debug_info: bool = False):
    """

    Nondeterministic if all_actions_tried == FALSE

    @param lowest_AIRIS_conf:
    @param lowest_AIRIS_conf_actions:
    @param oldest_actions:
    @param oldest_age: oldest observed age (smaller == older)
    @param all_actions_tried: True if all possible actions were in the set of applicable rules
    @param babbling_rates: dict of rates e.g. : {'curiosity': 0.5, 'exploit': 1.0, 'explore': 0.5}
    @param print_debug_info:
    @return:
    """

    # Calculate babbling rates, some are 0.5 if not all actions tried
    explore_curiosity_modulator = 1.0 if all_actions_tried else 0.5
    babbling_rates = {
        'curiosity': explore_curiosity_modulator,
        'exploit': 1.0,
        'explore': explore_curiosity_modulator
    }

    if lowest_AIRIS_conf == float("-inf"):
        exploit_babble = random.random() > babbling_rates['exploit']  # must always evaluate to false

    else:
        if all_actions_tried:
            exploit_babble = False
        else:
            exploit_babble = random.random() > 0.5

    if all_actions_tried:
        explore_babble = False
    else:
        # if all_actions_tried evaluates to False, else False 50% of time
        explore_babble = random.random() > 0.5

    code_path = "?"

    if lowest_conf_predicted_score_delta > 0.0:
        behavior = "ACHIEVE"
        action = lowest_AIRIS_conf_actions[0]
        plan = lowest_AIRIS_conf_actions
        code_path = "Score via Lowest AIRIS conf."
    elif oldest_age_predicted_score_delta > 0.0:
        behavior = "ACHIEVE"
        action = oldest_actions[0]
        plan = oldest_actions
        code_path = "Score via oldest age"
    else:
        if (lowest_AIRIS_conf >= 0.99 # we think we model the world fully
                or exploit_babble or len(lowest_AIRIS_conf_actions) == 0):
            # EXPLORE or BABBLE
            if (not exploit_babble and not explore_babble and oldest_age > 0.0 and
                    lowest_AIRIS_conf == 1.0 and
                    len(oldest_actions) != 0):
                behavior = "EXPLORE"
                code_path = "OLDEST_CELLS"
                action = oldest_actions[0]
                plan = oldest_actions
            else:
                behavior = "BABBLE"
                action = _choose_babble_action(available_actions=available_actions)
                code_path = "RANDOM_ACTION"
                plan = []
        else:  # lowest_AIRIS_conf must be < 0.99
            # ACHIEVE or CURIOUS
            # the oscillation path on world 2 runs through here always selecting lowest_AIRIS_conf_actions
            behavior = "ACHIEVE" if lowest_AIRIS_conf == float("-inf") else "CURIOUS"
            action = lowest_AIRIS_conf_actions[0]
            plan = lowest_AIRIS_conf_actions
            code_path = "Lowest AIRIS conf."




    if print_debug_info:
        print("behavior", behavior,
              "airis_score", lowest_AIRIS_conf,
              "favored_actions", lowest_AIRIS_conf_actions,
              "favored_actions_for_revisit", oldest_actions, "oldest_age>0", (oldest_age > 0.0),
              )

    print(behavior, nace.prettyprint.Prettyprint_AllActions(plan), "oldest_age:", oldest_age, "AIRIS Conf.:", lowest_AIRIS_conf, "Path:",code_path,
          "Lowest age reason:", lowest_AIRIS_conf_stopping_reason, "oldest_age_stopping_reason",
          oldest_age_stopping_reason)
    return behavior, action, plan


def _choose_babble_action(available_actions):
    if nace.world_module.drop in available_actions:
        available_actions += [nace.world_module.drop, nace.world_module.drop]
    return random.choice(available_actions)


def _create_explanation_graphs():
    """
    Hold some things constant, then produce graphs to try an explain how airis score based on synthetic inputs
    e.g.: all_actions_tried, age etc relate to each other
    @return:
    """
    # Calculate babbling rates
    all_results = {}
    num_repeats = 100
    number_airis_bins = 20
    description = ""

    for oldest_age in [10]:
        for all_actions_tried in [True, False]:
            results = []
            for i in range(number_airis_bins + 1):
                airis_score = i / float(number_airis_bins)
                action_type_counts = collections.defaultdict(int)
                behavior_counts = collections.defaultdict(int)

                for run in range(num_repeats):
                    # Determine behavior and action
                    behavior, action, plan = _determine_behavior_and_action(
                        airis_score,
                        lowest_AIRIS_conf_actions=["FA[0]", "FA[1]"],
                        oldest_actions=["FAFR[0]", "FAFR[1]"],
                        oldest_age=oldest_age,
                        available_actions=["FA[0]", "FA[1]", "FAFR[0]", "FAFR[1]"],
                        all_actions_tried=all_actions_tried,
                    )
                    behavior_counts[behavior] += 1
                    if action.find("FAFR[") > -1:
                        action_type_counts["favored_actions_for_revisit"] += 1
                    elif action.find("FA[") > -1:
                        action_type_counts["favored_actions"] += 1

                description = "oldest_age:" + str(oldest_age) + " all_actions_tried:" + str(all_actions_tried)
                results.append({"airis_score": airis_score,
                                "CURIOUS": behavior_counts["CURIOUS"],
                                "ACHIEVE": behavior_counts["ACHIEVE"],
                                "BABBLE": behavior_counts["BABBLE"],
                                "EXPLORE": behavior_counts["EXPLORE"],
                                "favored_actions_for_revisit": action_type_counts["favored_actions_for_revisit"],
                                "favored_actions": action_type_counts["favored_actions"],
                                "description": description,
                                'all_actions_tried': all_actions_tried,
                                'oldest_age': oldest_age
                                })
            print("________________ results:")
            print(results)
            all_results[description] = results
    print("________________ all results:")
    print(all_results)
    # return plan, action, rules_excluded, behavior


def _match_hypotheses(focus_set, oldworld, action, rules, old_agent: Type[Agent]):
    """
    Match hypotheses (rules) preconditions to the world, calculating how AIRIS-confident the prediction would be:
    ( called _MatchHypotheses in old system)

    @param focus_set:
    @param oldworld:
    @param action:
    @param rules:
    @return:
    """
    positionscores = dict([])
    highesthighscore = 0.0
    AttendPositions = set([])

    height, width = oldworld.get_height_width()
    for y in range(height):
        for x in range(width):
            if oldworld.get_char_at(y, x) in focus_set:
                # if oldworld[world_module.BOARD][y][x] in focus_set:
                AttendPositions.add((y, x))
                for rule in rules:
                    (precondition, consequence) = rule
                    action_score_and_preconditions = list(precondition)
                    for y_rel, x_rel, requiredstate in action_score_and_preconditions[
                                                       2:
                                                       ]:
                        AttendPositions.add((y + y_rel, x + x_rel))
    for y in range(height):
        for x in range(width):
            if (y, x) not in AttendPositions:
                continue
            scores = dict([])
            positionscores[(y, x)] = scores
            highscore = 0.0
            highscorerule = None
            for rule in rules:
                (precondition, consequence) = rule
                action_score_and_preconditions = list(precondition)
                values = action_score_and_preconditions[1]
                if action_score_and_preconditions[0] == action:
                    scores[rule] = 0.0
                else:
                    continue
                CONTINUE = False
                for i in range(len(values)):
                    if values[i] != old_agent.get_values_exc_score()[
                        i]:  # hmm should be doable? # 15th Aug, changed to be inclusive of score.
                        # 30 aug changed to be exlusive of score
                        CONTINUE = True
                if CONTINUE:
                    continue
                for y_rel, x_rel, requiredstate in action_score_and_preconditions[2:]:
                    if (
                            y + y_rel >= height
                            or y + y_rel < 0
                            or x + x_rel >= width
                            or x + x_rel < 0
                    ):
                        CONTINUE = True
                        break
                    if oldworld.get_char_at(y + y_rel, x + x_rel) == requiredstate:
                        scores[rule] += 1.0  # n matched_conds
                if CONTINUE:
                    continue
                scores[rule] /= len(precondition) - 2  # Q(r,c) - Match Quotient
                if scores[rule] > 0.0 and (
                        scores[rule] > highscore
                        or (
                                scores[rule] == highscore
                                and highscorerule is not None
                                and len(rule[0]) > len(highscorerule[0])
                        )
                ):
                    highscore = scores.get(rule, 0.0)
                    highscorerule = rule
            positionscores[(y, x)] = (scores, highscore, highscorerule)
            if highscore > highesthighscore:
                highesthighscore = highscore
    return (positionscores, highesthighscore)


def _rule_applicable(scores, highscore, highesthighscore, rule):  # called _RuleApplicable in old system
    """
    # Whether a rule is applicable: only if it matches better than not at all, and as well as the best matching rule

    @param scores:
    @param highscore:
    @param highesthighscore:
    @param rule:
    @return:
    """
    if highscore > 0.0 and scores.get(rule, 0.0) == highesthighscore:
        return True
    return False


def _predict_next_world_state(focus_set, oldworld, action, rules,
                              agent,
                              customGoal=None):  # called NACE_Predict in old system
    """
    Returns the world as predicted after 'action' is applied to 'oldWorld'.
    It does this by applying rules that have been previously learnt.

    How: Apply the move to the predicted world model whereby we use the learned rules to decide how grid elements might
    most likely change.

    :param focus_set: - Note: NOT updated by this routine.
    :param oldworld:
    :param action:
    :param rules:
    :param customGoal:
    :return:
    """
    newworld = copy.deepcopy(oldworld)
    newagent = copy.deepcopy(agent)
    used_rules_sumscore = 0.0
    used_rules_amount = 0
    score_delta = 0.0
    # score : the AIRIS confidence of the prediction
    (positionscores, highesthighscore) = _match_hypotheses(
        focus_set, oldworld, action, rules, agent
    )
    max_focus = None
    if len(focus_set) > 0:
        max_focus = max(focus_set, key=lambda k: focus_set[k])
    age = 0
    height, width = oldworld.get_height_width()
    for y in range(height):
        for x in range(width):
            if (y, x) not in positionscores:
                continue
            scores, highscore, rule = positionscores[(y, x)]
            # for rule in rules:
            if _rule_applicable(scores, highscore, highesthighscore, rule):
                # newworld[world_module.VALUES] = rule[1][3]    # values including score?
                newagent.set_values_inc_score(rule[1][3])  # new agent score should be considered delta
                newworld.set_char_at(y, x, rule[1][2])
                used_rules_sumscore += scores.get(rule, 0.0)
                used_rules_amount += 1
    current_time = newworld.get_newest_time()
    for y in range(height):
        for x in range(width):
            if (y, x) not in positionscores:
                continue
            if (
                    max_focus
                    and newworld.get_char_at(y, x) in focus_set
                    and newworld.get_char_at(y, x) == max_focus
            ):
                age = max(age, (current_time - newworld.get_time_at(y, x)))
    AIRIS_confidence = (
        used_rules_sumscore / used_rules_amount if used_rules_amount > 0 else 1.0
    )  # AIRIS confidence
    if AIRIS_confidence < 1.0:
        pass  # noop for breakpoint placement

    # but if the certainty predicted world has higher value, then set prediction score to the best it can be
    if (newagent.get_score() > 0.0 and AIRIS_confidence == 1.0):  # newagent score should be considered as a delta here
        AIRIS_confidence = float("-inf")  # set this to be where we would explore
        score_delta = newagent.get_score()

    # while if the certainty predicted world has lower value, set prediction score to the worst it can be
    if (newagent.get_score() < 0.0 and AIRIS_confidence == 1.0):
        AIRIS_confidence = float("inf")  # i.e. Flag to tell calling code avoid! avoid! # needs to be re-worked
        score_delta = newagent.get_score()  # newagent score should be considered as a delta here

    if (customGoal and customGoal(newworld)):
        AIRIS_confidence = float("-inf")  # set this to be where we would explore
        score_delta = 1.0
        newagent.set_score(score_delta)  # newagent score should be considered as a delta here

    # l = max(len(agent.get_values_inc_score()), len(newagent.get_values_inc_score()))
    # value_deltas = [0] * l
    # for i in range(l):
    #     value_deltas[i] = newagent.get_values_inc_score()[i] - agent.get_values_inc_score()[i]

    return (newworld,
            AIRIS_confidence,
            age,
            newagent.get_values_inc_score(),  # delta of all values including score
            score_delta)


def _add_to_adjacent_set(adjacent_change_sets: list, newEntry: tuple, MaxCapacity: int, CanCreateNewSet: bool,
                         maximum_distance: int = 1):  # called _AddToAdjacentSet in the old system
    """


    If there are no entries in the adjacent_change_sets create a new one with the new point.
    If the new point if not adjacent to any of the existing sets create a new set containing it.
    If the new point is adjacent to a point in the existing sets, add it to that set if
              there are less than 3 points in it.

    @param adjacent_change_sets:
    @param newEntry:
    @param MaxCapacity:
    @param CanCreateNewSet:
    @param maximum_distance:
    @return:
    """
    (y, x) = newEntry
    AdjacentToAnySet = False
    for consideredSet in adjacent_change_sets:
        consideredSetFrozen = copy.deepcopy(consideredSet)
        for ys, xs in consideredSetFrozen:
            if abs(y - ys) + abs(x - xs) <= maximum_distance:
                if len(consideredSet) < MaxCapacity:
                    consideredSet.add(newEntry)
                AdjacentToAnySet = True
    if not AdjacentToAnySet and CanCreateNewSet:
        adjacent_change_sets.append({newEntry})


def _is_presently_observed(Time, world, y, x):
    """
    # Whether the grid cell has been observed now (not all have been, due to partial observability)

    @param Time:
    @param world:
    @param y:
    @param x:
    @return:
    """
    diff = Time - world.times[y][x]
    return diff == 0  # TODO this logic should be in world object


def _build_change_sets(
        focus_set,
        oldworld,
        action,
        newworld,
        predictedworld,
        unobserved_code,
        object_count_threshold=1
):
    """
    calc value_counts : counts each board value has been seen

    @param focus_set:
    @param oldworld:
    @param action:
    @param newworld:
    @param predictedworld:
    @param object_count_threshold:
    @return:
    """
    # Keep track of cell type counts - count of each cell type
    # in world keyed by the chars in the world map
    value_counts = dict([])
    height, width = oldworld.get_height_width()
    for y in range(height):
        for x in range(width):
            val = oldworld.get_char_at(y, x)
            if val not in value_counts:
                value_counts[val] = 1
            else:
                value_counts[val] += 1
    # Update focus_set based on unique values and unique changing values.
    # focus_set : keyed by the char in a map location.
    #         Value == the number of time steps this value has been one that has changed (and there was only 1 of them)
    # if the world is huge, is this realistic, or work out from current location till N are found ...
    #
    pass
    # only consider cells that were observed in both newworld and oldworld in the most recent upadate of each,
    # if there is only 1 cell of this type (why?) add this, or increment this in the focus set.
    # Optimisation : Code could be changed to pull this set of locations from the worlds utilizing fast np operations.
    current_time = newworld.get_newest_time()
    for y in range(height):
        for x in range(width):
            # skip unobserved cells
            if not _is_presently_observed(
                    current_time, newworld, y, x
            ) and not _is_presently_observed(current_time - 1, newworld, y, x):
                continue
            val = oldworld.get_char_at(y, x)  # value in the map
            if val == "O":
                pass
            # if val in value_counts.keys() and val not in focus_set:  # if 'val' is new, init focus set
            #     focus_set[val] = 0
            # if oldworld.get_char_at(y, x) != newworld.get_char_at(y, x):  # this cell changed away from value 'val'
            #     # used to check for uniquiness - was this just an oversight? - why check for uniqueness? this makes no sense. # dv-5/Aug # Patrick Why
            #     if val not in focus_set:
            #         focus_set[val] = 1
            #     else:
            #         focus_set[val] += 1  # == the number of times this cell type has been seen to change
            #
            #
            # if 1 <= value_counts[
            #     val] <= object_count_threshold and val not in focus_set:  # if this map type is unique and not in focus set # Patrick Why
            #     focus_set[val] = 0
            if value_counts[val] == 1 and val not in focus_set:  # if 'val' is new, init focus set
                focus_set[val] = 0
            if oldworld.get_char_at(y, x) != newworld.get_char_at(y, x):  # only when it comes into view first time
                if 1 <= value_counts[val] <= object_count_threshold:  # unique - why check for uniqueness?
                    # this makes no sense. # dv-5/Aug # Patrick said implementation detail, it meant that it would only
                    # trigger if there was 1 new value within visible window
                    focus_set[val] += 1  # can happen when already set to 0, or over multiple time steps.
                    # because we may focus over multiple time frames?

    # create adjacent change sets
    adjacent_change_sets = []
    for y in range(height):
        for x in range(width):
            # skip cells not observed in either this, or the last times step.
            if _is_presently_observed(
                    current_time, newworld, y, x
            ) and _is_presently_observed(current_time - 1, newworld, y, x):
                if (  # if different and not observed code
                        oldworld.get_char_at(y, x) != newworld.get_char_at(y, x)
                        and oldworld.get_char_at(y, x) != unobserved_code
                ):
                    # moving into the cliff or off the edge of the board makes no change, so this code does not detect it.
                    if oldworld.get_char_at(y,
                                            x) == 'C':
                        pass  # debugging breakpoint
                    # mutate changesets adding values to it.
                    _add_to_adjacent_set(
                        adjacent_change_sets, (y, x), MaxCapacity=3, CanCreateNewSet=True, maximum_distance=1
                        # MaxCapacity=3 as 3 things need to have changed for more complex rules to be induced.
                        # TODO 3 and 1 are magic numbers.
                    )
    changesetslen = len(adjacent_change_sets)  # the length of change set only
    changeset0len = 0  # length of changeset 0 (the first changeset) Possibly this should be the max changeset length?
    if changesetslen > 0:
        changeset0len = len(
            adjacent_change_sets[0]
        )  # temporary fix: 3 things need to have changed at least to allow for the more complex rules to be induced
    # Add prediction mismatch entries to adjacent change set entry (using newworld for observation times)
    pass
    for y in range(height):
        for x in range(width):
            # skip unobserved cells
            if not _is_presently_observed(
                    current_time, newworld, y, x
            ) and not _is_presently_observed(current_time - 1, newworld, y, x):
                continue
            if (
                    predictedworld
                    and predictedworld.get_char_at(y, x) != newworld.get_char_at(y, x)
                    and oldworld.get_char_at(y, x) != unobserved_code
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y, x), MaxCapacity=2, CanCreateNewSet=True, maximum_distance=1
                    # MaxCapacity is one less than it was. hmmm...
                )
    # if there was a change next to a focus set element (spatial dependency) add it to the changeSet
    chgsets = copy.deepcopy(adjacent_change_sets)
    for changeset in chgsets:
        for y, x in changeset:
            if (
                    (action == nace.world_module.left or action == nace.world_module.right)
                    and x > 0  # assumes there will be a wall round the board
                    and newworld.get_char_at(y, x - 1) in focus_set
                    and oldworld.get_char_at(y, x - 1) != unobserved_code
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y, x - 1), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            if (
                    (action == nace.world_module.left or action == nace.world_module.right)
                    and x < width - 1  # assumes there will be a wall round the board
                    and newworld.get_char_at(y, x + 1) in focus_set
                    and oldworld.get_char_at(y, x + 1) != unobserved_code
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y, x + 1), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            if (
                    (action == nace.world_module.up or action == nace.world_module.down or action == nace.world_module.drop)
                    and y > 0  # assumes there will be a wall round the board
                    and newworld.get_char_at(y - 1, x) in focus_set
                    and oldworld.get_char_at(y - 1, x) != unobserved_code
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y - 1, x), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
            if (
                    (action == nace.world_module.up or action == nace.world_module.down)
                    and y < height - 1  # assumes there will be a wall round the board
                    and newworld.get_char_at(y + 1, x) in focus_set
                    and oldworld.get_char_at(y + 1, x) != unobserved_code
            ):
                _add_to_adjacent_set(
                    adjacent_change_sets, (y + 1, x), MaxCapacity=3, CanCreateNewSet=False, maximum_distance=1
                )
    # Change sets are now built ready to be used.
    print("Adjacent change sets have been built:", adjacent_change_sets)
    return focus_set, adjacent_change_sets, changeset0len


def _observe(
        focus_set,
        rule_evidence,
        oldworld,
        action,
        newworld,
        oldrules,
        oldnegrules,
        predictedworld,
        pre_action_agent,
        ground_truth_post_action_agent,
        unobserved_code,
        object_count_threshold,
):
    """
    Extract new rules from the observations by looking only for observed changes and prediction-observation mismatches

    @param focus_set: Set of focus points
    @param rule_evidence:  dict[rule:(+ve evidence, -ve evidence)]
    @param oldworld: State of the world before the action
    @param action: Action taken
    @param newworld: State of the world after the action
    @param oldrules: Existing rules
    @param oldnegrules: Existing negative rules
    @param predictedworld: Predicted state of the world after the action
    @param pre_action_agent: Agent state before the action
    @param ground_truth_post_action_agent: Ground truth agent state after the action
    @param unobserved_code: Code representing unobserved states
    @param object_count_threshold: Threshold for object count
    @return: Updated focus_set, rule_evidence, new_rules, new_negrules
    """
    new_rules = copy.deepcopy(oldrules)
    new_negrules = copy.deepcopy(oldnegrules)

    if ground_truth_post_action_agent.get_score() >= 1:
        pass  # used to place breakpoint for debugging. TODO remove

    focus_set, adjacent_change_sets, changeset0len = _build_change_sets(
        focus_set,
        oldworld,
        action,
        newworld,
        predictedworld,
        unobserved_code,
        object_count_threshold=object_count_threshold
    )

    # Build rules based on changes and prediction-observation mismatches
    # Algo: Compare all to all others,
    #
    for changeset in adjacent_change_sets:
        for y1_abs, x1_abs in changeset:
            action_values_precondition = [action,
                                          tuple(pre_action_agent.get_values_exc_score())]  # values excluding score
            preconditions = []
            CONTINUE = False
            for y2_abs, x2_abs in changeset:
                (y2_rel, x2_rel) = (y2_abs - y1_abs, x2_abs - x1_abs)  # relative to changeset 1
                condition = (y2_rel, x2_rel, oldworld.get_char_at(y2_abs, x2_abs))
                if (
                        oldworld.get_char_at(y2_abs, x2_abs) == unobserved_code
                ):  # NECESSARY FOR EPISODE RESET ONLY
                    CONTINUE = True  # old world at this location was unobserved. So no info here, skip?
                    break
                if Hypothesis_ValidCondition(condition):  # close by (0-2 in distance)
                    preconditions.append(condition)
                    # if current location in old_board has the same object on it in the newworld, we can skip forward
                    if y2_rel == 0 and x2_rel == 0:
                        if (oldworld.get_char_at(y2_abs, x2_abs)
                                == newworld.get_char_at(y1_abs, x1_abs)):
                            #
                            CONTINUE = True
                            break
            if CONTINUE:
                continue  # skip forward to next record
            preconditions = sorted(preconditions)
            for pr in preconditions:
                action_values_precondition.append(pr)
            rule = (
                tuple(action_values_precondition),
                (
                    0,
                    0,
                    newworld.get_char_at(y1_abs, x1_abs),
                    tuple(
                        [post_v - pre_v for (post_v, pre_v) in (
                            zip(ground_truth_post_action_agent.get_values_inc_score(),
                                pre_action_agent.get_values_inc_score()))]  # get the delta of each value inc score
                    ),
                ),
            )
            if len(preconditions) >= 2 and (
                    changeset0len == 3 or len(preconditions) <= 2
            ):
                rule_evidence, new_rules = Hypothesis_Confirmed(  # Mutates returned copy of rule_evidence and ruleset
                    focus_set, rule_evidence, new_rules, new_negrules, rule
                    # note newrules can be mutated in this routine.
                )
        break  # speedup (dv-this looks odd, why have the outer loop if we always break?)
        # hmm, related to the continue flag.

    # if rule conditions are only partly met or the predicted outcome is different from observed,
    # build a specialized rule which has the precondition and conclusion corrected!
    max_focus = None
    if len(focus_set) > 0:
        max_focus = max(focus_set, key=lambda m: focus_set[m])
    (positionscores, highesthighscore) = _match_hypotheses(  # scores == Match Quotient
        focus_set, oldworld, action, new_rules, pre_action_agent
    )

    height, width = oldworld.get_height_width()
    current_time = newworld.get_newest_time()

    for y in range(height):
        for x in range(width):
            if (y, x) not in positionscores:
                continue
            if (
                    not _is_presently_observed(current_time, newworld, y, x)
                    and oldworld.get_char_at(y, x) != max_focus
                    and not (
                    newworld.get_char_at(y, x) == unobserved_code and oldworld.get_char_at(y, x) != unobserved_code)
            ):
                continue
            scores, highscore, rule = positionscores[(y, x)]
            # for rule in oldrules:
            if _rule_applicable(scores, highscore, highesthighscore, rule):
                if rule[1][2] != newworld.get_char_at(y, x):
                    (precondition, consequence) = rule
                    action_score_and_preconditions = list(precondition)
                    # values = action_score_and_preconditions[1]  # dv commented out as not used 3/aug/2024
                    corrected_preconditions = []
                    CONTINUE = False
                    has_focus_set_condition = False  # TODO!!!
                    for y_rel, x_rel, requiredstate in action_score_and_preconditions[
                                                       2:
                                                       ]:
                        if (
                                y + y_rel >= height
                                or y + y_rel < 0
                                or x + x_rel >= width
                                or x + x_rel < 0
                        ):
                            CONTINUE = True
                            break
                        if oldworld.get_char_at(y + y_rel, x + x_rel) == max_focus:
                            has_focus_set_condition = True
                        if oldworld.get_char_at(y + y_rel, x + x_rel) == unobserved_code:
                            CONTINUE = True
                            break
                        corrected_preconditions.append(
                            (y_rel, x_rel, oldworld.get_char_at(y + y_rel, x + x_rel))
                        )
                    corrected_preconditions = sorted(corrected_preconditions)
                    if CONTINUE or not has_focus_set_condition:
                        continue
                    rule_new = (
                        tuple(
                            [
                                action_score_and_preconditions[0],
                                action_score_and_preconditions[1],
                            ]
                            + corrected_preconditions
                        ),
                        tuple(
                            [
                                rule[1][0],
                                rule[1][1],
                                newworld.get_char_at(y, x),
                                tuple(
                                    [post_v - pre_v for (post_v, pre_v) in (
                                        zip(ground_truth_post_action_agent.get_values_inc_score(),
                                            pre_action_agent.get_values_inc_score()))]
                                    # get the delta of each value inc score
                                ),
                            ]
                        ),
                    )
                    # print("RULE CORRECTION ", y, x, xy_loc, worldchange);
                    # Prettyprint_rule(rule);
                    # Prettyprint_rule(rule_new)
                    rule_evidence, new_rules = Hypothesis_Confirmed(
                        # Mutates returned copy of rule_evidence and ruleset
                        focus_set, rule_evidence, new_rules, new_negrules, rule_new
                    )
                    break
    # Crisp match: Add negative evidence for rules which prediction contradicts observation (in a classical AIRIS
    # implementation restricted to deterministic worlds: this part would remove contradicting rules from the
    # rule set and would ensure they can't be re-induced)
    for y in range(height):
        for x in range(width):
            if (
                    not _is_presently_observed(current_time, newworld, y, x)
                    and oldworld.get_char_at(y, x) != max_focus
                    and not (
                    newworld.get_char_at(y, x) == unobserved_code and oldworld.get_char_at(y, x) != unobserved_code)
            ):
                continue
            for (
                    rule
            ) in (
                    oldrules
            ):  # find rules which don't work, and add negative evidence for them (classical AIRIS:
                # remove them and add them to newnegrules)
                (precondition, consequence) = rule
                action_valsExScore_and_preconditions = list(precondition)
                valsExScore = action_valsExScore_and_preconditions[1]
                CONTINUE = False
                if action_valsExScore_and_preconditions[0] != action:  # rule did not apply
                    continue
                for i in range(len(valsExScore)):
                    if (
                            valsExScore[i] != pre_action_agent.get_values_exc_score()[i]  #
                    ):  # value didn't match, rule did not apply
                        CONTINUE = True
                        break
                for y_rel, x_rel, requiredstate in action_valsExScore_and_preconditions[2:]:
                    if (
                            y + y_rel >= height
                            or y + y_rel < 0
                            or x + x_rel >= width
                            or x + x_rel < 0
                    ):
                        CONTINUE = True
                        break
                    if oldworld.get_char_at(y + y_rel, x + x_rel) != requiredstate:
                        CONTINUE = True
                        break
                if CONTINUE:
                    continue
                if rule[1][3][0] != ground_truth_post_action_agent.get_score() - pre_action_agent.get_score():
                    rule_evidence, new_rules, new_negrules = Hypothesis_Contradicted(
                        # Mutates returned copy of rule_evidence
                        rule_evidence, new_rules, new_negrules, rule
                    )  # score increase did not happen
                    continue
                num_values = max(len(rule[1][3]), len(ground_truth_post_action_agent.get_values_inc_score()))
                for k in range(1, num_values):  # wrong value (not score) prediction (we start at index 1)
                    agent_value_k = ground_truth_post_action_agent.get_values_inc_score()[k]
                    rule_value_k = rule[1][3][k]
                    if rule_value_k != agent_value_k:
                        rule_evidence, new_rules, new_negrules = Hypothesis_Contradicted(
                            # Mutates returned copy of rule_evidence
                            rule_evidence, new_rules, new_negrules, rule
                        )
                        CONTINUE = True
                        break
                if CONTINUE:
                    continue
                if rule[1][2] != newworld.get_char_at(y, x):
                    rule_evidence, new_rules, new_negrules = Hypothesis_Contradicted(
                        # Mutates returned copy of rule_evidence
                        rule_evidence, new_rules, new_negrules, rule
                    )
    return focus_set, rule_evidence, new_rules, new_negrules


def nacev3_predict_and_observe(  # called _predict_and_observe in v1
        time_counter,
        focus_set,
        rule_evidence,
        xy_loc,  # pre action agent location note (0,0) is top left
        pre_action_world,  # pre action internal world model (partially observed) before last_action is applied
        rulesin,  # used in part1 and part2
        negrules,
        plan,
        last_action,
        rulesExcluded,
        post_action_ground_truth_world,  # i.e. this is the post action world model
        pre_action_agent,  # pre action agent
        ground_truth_post_action_agent,  # post action agent
        unobserved_code,  # value which indicates we can not see the true value of the cell
        object_count_threshold=1,  # how unique values must be before they are considered. was 1, i.e. unique
        print_out_world_and_plan=True):
    """
    Find difference between the actual world, and the predicted world.

    @param time_counter:
    @param focus_set:
    @param rule_evidence:
    @param xy_loc:
    @param pre_action_world:
    @param rulesin:
    @param negrules:
    @param plan:
    @param last_action:
    @param rulesExcluded:
    @param post_action_ground_truth_world:
    @param pre_action_agent
    @param ground_truth_post_action_agent
    @param print_out_world_and_plan:
    @return:
    """
    if print_out_world_and_plan:
        print("-- Predict and Observe Start ---")
    rules = copy.deepcopy(rulesin)
    simulated_world_original = copy.deepcopy(pre_action_world)
    simulated_world_post_action = copy.deepcopy(pre_action_world)
    xy_location_t_minus_1 = simulated_world_original.get_agent_xy_location()
    modified_count, _ = simulated_world_post_action.update_world_from_ground_truth(
        time_counter,
        post_action_ground_truth_world,
        [xy_loc, xy_location_t_minus_1]
    )

    r = simulated_world_post_action.get_board_char_counts()
    if "x" in r:
        if r["x"] != 1:
            pass

    stayed_the_same = modified_count == 0

    rules, rulesExcluded = Hypothesis_BestSelection(
        rules,
        rulesExcluded,
        rule_evidence,
        stayed_the_same)  # Step 2 # Mutates rules and rulesExcluded by adding and removing items

    rules = copy.deepcopy(rulesin)
    predicted_world, _, __, values, predicted_score_delta = _predict_next_world_state(
        focus_set, copy.deepcopy(simulated_world_original), last_action, rules, pre_action_agent
    )

    # Extract new rules from the observations by looking only for observed changes and prediction-observation mismatches
    focus_set, rule_evidence, newrules, newnegrules = _observe(
        focus_set,
        rule_evidence,
        simulated_world_original,
        last_action,
        simulated_world_post_action,
        rules,
        negrules,
        predicted_world,
        pre_action_agent,
        ground_truth_post_action_agent,
        unobserved_code,
        object_count_threshold
    )
    used_rules = copy.deepcopy(newrules)
    for rule in rulesExcluded:  # add again so we won't lose them
        newrules.add(rule)

    lastplanworld = copy.deepcopy(simulated_world_original)
    planworld = copy.deepcopy(predicted_world)
    for i in range(1, len(plan)):
        lastplanworld = copy.deepcopy(planworld)
        planworld, _, __, ___, ____ = _predict_next_world_state(
            focus_set, copy.deepcopy(planworld), plan[i], rules, pre_action_agent
        )

    if print_out_world_and_plan:
        post_action_ground_truth_world.multiworld_print(
            [
                {"Caption": "Internal model:\ndt=-1",
                 "World": simulated_world_original,
                 "Color": nace.color_codes.color_code_white_on_blue},

                {"Caption": "Predicted\ndt=0\n==GT?",
                 "World": predicted_world,
                 "Color": nace.color_codes.color_code_white_on_blue},

                {"Caption": f"Ground Truth \ndt=0\nt={time_counter} beliefs={len(rules)}:",
                 "World": post_action_ground_truth_world,
                 "Color": nace.color_codes.color_code_white_on_black},

                {"Caption": "End Prediction\nsteps=" + str(len(plan)) + "\n" + str(Prettyprint_AllActions(plan)),
                 "World": planworld,
                 "Color": nace.color_codes.color_code_white_on_red},
            ]
        )
    if print_out_world_and_plan:
        print("-- Predict and Observe END ---")

    return (
        used_rules,
        focus_set,
        rule_evidence,
        simulated_world_post_action,
        newrules,
        newnegrules,
        values,
        lastplanworld,
        planworld,
        stayed_the_same
    )

    # return plan, action, rulesExcluded, behavior, simulated_world_staued_the_same_last_time_new_data_copied_in


if __name__ == "__main__":
    _create_explanation_graphs()



"""
 * The MIT License
 *
 * Copyright (c) 2024 Patrick Hammer
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 """
import copy
import random
import sys
from copy import deepcopy

import nace.color_codes
import nace.world_module_numpy
import nace.world_module

# _challenge_input = ""
# for arg in sys.argv:
#     if arg.startswith("world="):
#         _challenge_input = arg.split("world=")[1]
# # Description for world choice
# if "manual" in sys.argv:
#     if _challenge_input == "":
#         print("Enter one of 1-9 to try a world:")
# elif "no_world" in sys.argv:
#     _challenge_input = "0"
# else:
#     print(
#         'Food collecting (1), cup on table challenge (2), doors and keys (3), food collecting with moving object (4), pong (5), bring eggs to chicken (6), soccer (7), shock world (8), interactive world (9), input "1", "2", "3", "4", "5", "6", or "7":')
# if _challenge_input == "":
#     _challenge = input()
# else:
#     _challenge = _challenge_input

# print('Slippery ground y/n (n default)? Causes the chosen action to have the consequence of another action in 10% of cases.')
_slippery = "slippery" in sys.argv
_isWorld5 = False
_isWorld9 = False


def getIsWorld9():
    return _isWorld9


def set_world(_challenge_number):
    global _challenge
    _challenge = _challenge_number


def get_height_width(world):
    """ return the height and width of the board, so that callers do not need to know it's internal structure"""
    if len(world[BOARD]) > 0:
        height, width = (len(world[BOARD]), len(world[BOARD][0]))
    else:
        height, width = (0, 0)
    return height, width


def is_location_on_board(world, xy_loc):
    location_col, location_row = xy_loc
    height, width = (len(world[BOARD]), len(world[BOARD][0]))
    return location_row >= 0 and location_row < height and location_col >= 0 and location_col < width


# Whether there is a cup on the table (user-given goal demo)
def World_CupIsOnTable(world):
    if type(world) is nace.world_module_numpy.NPWorld:
        world_list = world.to_string_list()
    else:
        world_list = world

    height, width = get_height_width(world_list)
    for x in range(width):
        for y in range(height - 1):
            if world_list[BOARD][y + 1][x] == 'T' and world_list[BOARD][y][x] == 'u':
                return True
    return False


World_objective = None
# World states:

env = None
dir_right = 0
dir_down = 1
dir_left = 2
dir_up = 3
action_left = 0
action_right = 1
action_forward = 2
action_pick = 3
action_drop = 4
action_toggle = 5
lastimage = None

gwv_location_xy = (2, 4)  # never used in this file?
gwv_direction = None  # never assigned from main.py
gwv_last_reward = None
gwv_has_reset = 0
gwv_reset_score = 0
gwv_accumulated_score = 0
gwv_step = 0
gwv_episode_id = 0

run_id = -1
episodes = -1
max_steps = 1000

for arg in sys.argv:
    if arg.startswith("runid="):
        run_id = int(arg.split("runid=")[1])
    if arg.startswith("episodes="):
        episodes = int(arg.split("episodes=")[1])


def minigrid_digest(state):
    global gwv_direction, loc, lastimage, gwv_last_reward, gwv_step
    gwv_step += 1
    # print(state[0]["direction"]); exit(0)
    gwv_direction = state[0]["direction"]
    # print(state[0]['image'].shape); exit(0)
    loc = env.agent_pos
    lastimage = state[0]["image"]
    if len(state) == 5 and (gwv_last_reward == 0 or gwv_last_reward == None) and state[1] != 0:
        gwv_last_reward = state[1]
        reward = state[1]


VIEWDISTX, VIEWDISTY = (3, 2)
COFFEEMACHINE, WALL, ROBOT, CUP, FOOD, BATTERY, FREE, TABLE, GOAL, KEY, DOOR, ARROW_DOWN, ARROW_UP, BALL, EGG, EGGPLACE, CHICKEN, SBALL, SHOCK = \
    ('G', 'o', 'x', 'u', 'f', 'b', ' ', 'T', 'H', 'k', 'D', 'v', '^', 'c', 'O', '_', '4', '0', 'z')


def get_world_str(_challenge, max_steps=1000):
    global world_str

    global _isWorld9
    global _isWorld5
    global isMinigridWorld
    global World_objective

    World_objective = None

    # The worlds:
    world_str = """
    oooooooooooo
    o   o   f  o
    o          o
    o   oooooooo
    o x        o
    o       u  o
    oooooooooooo
    """
    _world2 = """
    oooooooooooo
    o          o
    o   u      o
    o     ooooTo
    o x        o
    o          o
    oooooooooooo
    """
    _world3 = """
    oooooooooooo
    o   k o   ko
    o     D b  o
    o     oooooo
    o x   D b  o
    o     o   ko
    oooooooooooo
    """
    _world4 = """
    oooooooooooo
    o  vo   f  o
    o          o
    o   oooooooo
    o x        o
    o       u  o
    oooooooooooo
    """
    _world5 = """
    oooooooooooo
    oo          
    oo          
    oo         c
    oox         
    oo          
    oooooooooooo
    """
    _world6 = """
    oooooooooooo
    o O  4  O  o
    o          o
    o          o
    o x  O  O  o
    o          o
    oooooooooooo
    """
    _world7 = """
    oooooooooooo
    o          o
    o  0 0     o
    o       H  o
    o x0 0     o
    o          o
    oooooooooooo
    """
    _world8 = """
    oooooooooooo
    o     zzzzzo
    o f z z    o
    o   z z zz o
    o x z z z  o
    o   z   z  o
    oooooooooooo
    """
    _world9 = """
    oooooooooooo
    o          o
    o     u    o
    o  T     G o
    o x   H    o
    o          o
    oooooooooooo
    """
    _world_empty = """











    """
    world_test_1 = """
    oooooo
    o   fo
    ox   o
    oooooo
    """

    if "t1" == _challenge:
        world_str = world_test_1
    if "2" == _challenge:
        world_str = _world2
        World_objective = World_CupIsOnTable
    if "3" == _challenge:
        world_str = _world3
    if "4" == _challenge:
        world_str = _world4
    if "5" == _challenge:
        world_str = _world5
        _isWorld5 = True
        global VIEWDISTX
        global VIEWDISTY
        VIEWDISTX, VIEWDISTY = (4, 3)

    if "6" == _challenge:
        world_str = _world6
    if "7" == _challenge:
        world_str = _world7
    if "8" == _challenge:
        world_str = _world8
    if "9" == _challenge:
        world_str = _world9
        _isWorld9 = True

    worldstr = "MiniGrid-DoorKey-8x8-v0"
    if "10" == _challenge:
        worldstr = "MiniGrid-Empty-8x8-v0"
    if "11" == _challenge:
        worldstr = "BabyAI-GoToRedBallNoDists-v0"
    if "12" == _challenge:
        worldstr = "MiniGrid-DistShift2-v0"
    if "13" == _challenge:
        worldstr = "MiniGrid-LavaGapS7-v0"
    if "14" == _challenge:
        worldstr = "MiniGrid-SimpleCrossingS11N5-v0"
    if "15" == _challenge:
        worldstr = "MiniGrid-LavaCrossingS11N5-v0"
    if "16" == _challenge:
        worldstr = "MiniGrid-Unlock-v0"
    if "17" == _challenge:
        worldstr = "MiniGrid-DoorKey-8x8-v0"
    if "18" == _challenge:
        worldstr = "MiniGrid-UnlockPickup-v0"
    if "19" == _challenge:
        worldstr = "MiniGrid-BlockedUnlockPickup-v0"

    isMinigridWorld = _challenge.isdigit() and int(_challenge) >= 10  # Minigrid challenges start at that index
    # if isMinigridWorld:  # "9" in _challenge:
    #     from minigrid.wrappers import *
    #
    #     _isWorld5 = False  # TODO
    #     if "nominigrid" not in sys.argv:
    #         env = gym.make(worldstr, render_mode='human', max_steps=max_steps)
    #     else:
    #         env = gym.make(worldstr, max_steps=max_steps)
    #     observation_reward_and_whatever = env.reset()
    #     minigrid_digest(observation_reward_and_whatever)
    #     print("Observation:", observation_reward_and_whatever)
    #     if "nominigrid" not in sys.argv:
    #         env.render()
    #     world_str = _world_empty
    #     loc = env.agent_pos

    if _isWorld5:
        set_full_action_list(list(set([up, down, right])))
    if isMinigridWorld and int(_challenge) >= 16:
        set_full_action_list(list(set([left, right, up, down, drop, pick, drop, toggle])))

    # strip leading and trailing spaces that were introduced by python tabs.
    t_world_str = world_str.split("\n")
    t_world_str_list = [s.strip() for s in t_world_str if len(s.strip()) > 0]

    if _isWorld5:
        t_world_str_list = [t.ljust(len(t_world_str_list[0]), " ") for t in t_world_str_list]

    for i, str in enumerate(t_world_str_list):
        assert len(str) == len(t_world_str_list[0])

    world_str = "\n".join(t_world_str_list)
    return world_str


#
# World is a tuple:
# (
#   Array of char that is a map,
#   tuple : ( Score?, Value? )
#   Array of int, how long ago the map was observed at that location, (if never observed -inf)
# )


BOARD, VALUES, TIMES = (0, 1, 2)


def build_initial_world_object(_challenge, unobserved_code):
    world_str = get_world_str(_challenge)
    world_obj = [[[*x] for x in world_str.split("\n")], tuple([0, 0])]
    _height, _width = get_height_width(world_obj)
    world_obj.append([[float("-inf") for i in range(_width)] for j in range(_height)])

    observed_world = [
        [[unobserved_code for x in world_obj[BOARD][i]] for i in
         range(len(world_obj[BOARD]))],
        world_obj[VALUES],
        world_obj[TIMES],
    ]
    return world_obj, observed_world, _height, _width


def build_blank_world_surrounded_by_wall(height, width, wall_char="\u2588", free_char='_', agent_char='x', agent_col=1,
                                         agent_row=1):
    result = ""
    for c in range(height):
        line = ""
        for r in range(width):
            if r == agent_row and c == agent_col:
                line += agent_char
            else:
                if r in [0, width - 1] or c in [0, height - 1]:
                    line += wall_char
                else:
                    line += free_char
        result += line + "\n"
    return result


def extract_agent_location_from_board(board, location_indicator_char='x'):
    """
    Extrat the location of an item on the board. Assumes uniqueness
    @param board:
    @param location_indicator_char:
    @return:
    """
    location_row = 0
    location_column = 0
    # check the sizes are correct, and see where location marker is
    for row_number in range(len(board)):
        if row_number > 0:
            assert len(board[0]) == len(board[row_number])
        # check for location marker
        if location_indicator_char in board[row_number]:
            location_row = row_number
            location_column = board[row_number].index(location_indicator_char)
    return (location_column, location_row)


def convert_gymnasium_world_to_nace(gym_world: str, strip_blanks=True, location_indicator_char='x', state_source=None,
                                    add_surrounding_walls=True, wall_code='W'):
    """

    @param gym_world: str , each \n indicates a new row, white space is stripped. agent location indicated by  location_indicator_char
    @param strip_blanks:
    @param location_indicator_char:  char representing the agent
    @param last_world: if not none, state and times seen copied from this world
    @param add_walls_right_and_bottom:
    @param wall_code:
    @return:
    """
    board = []  # list of rows. Each row is a list of char.
    for row_number, row_str in enumerate(gym_world.split("\n")):
        if len(row_str.strip()) > 0:
            if strip_blanks:
                board.append([c for c in list(row_str) if c != " "])
            else:
                board.append(list(row_str))
    if add_surrounding_walls:
        for row in board:
            row.append(wall_code)
            row.insert(0, wall_code)
        extra_line = [wall_code for _ in row]
        board.append(extra_line)
        board.insert(0, extra_line)

    BOARD, VALUES, TIMES = (0, 1, 2)

    if state_source is None:
        world_state = tuple([0, 0])
        height, width = (len(board), len(board[0]))
        world_observed_at = [[float("-inf") for i in range(width)] for j in range(height)]
    else:
        world_state = copy.deepcopy(state_source[VALUES])
        world_observed_at = copy.deepcopy(
            state_source[TIMES])  # this should never be used, here to speed conversion of code.

    nace_world = [board, world_state, world_observed_at]

    xy_location = extract_agent_location_from_board(board, location_indicator_char=location_indicator_char)

    return nace_world, xy_location  # NOTE location is xy format


# Move operations to move the agent in the world:
def left(loc):
    return (loc[0] - 1, loc[1])


def right(loc):
    if _isWorld5:
        return loc
    return (loc[0] + 1, loc[1])


def up(loc):
    return (loc[0], loc[1] - 1)


def down(loc):
    return (loc[0], loc[1] + 1)


def pick(loc):
    return loc


def drop(loc):
    return loc


def toggle(loc):
    return loc


gwv_lastseen = set([])


def help(action):
    if action != drop:
        minigrid_digest(env.step(action_toggle))
        minigrid_digest(env.step(action_pick))


def cntEntry(world, VAL):
    cnt = 0
    height, width = get_height_width(world)
    for y in range(height):
        for x in range(width):
            if world[BOARD][y][x] == VAL:
                cnt += 1
    return cnt


def World_Move(local_location, world, action, external_reward_for_last_action, annomate_arrows_prob=0.25):
    """

    Applies the effect of the movement operations, considering how different grid cell types interact with each other.
    Contains specific, hard coded logic, that the agent is expected to learn, or work around.

    @param local_location: assume xy location?
    @param world:
    @param action:
    @return:
    """
    global gwv_lastseen  # does not seem to need to be global
    global gwv_last_reward, gwv_has_reset, gwv_reset_score, gwv_accumulated_score, gwv_step, gwv_episode_id
    gwv_lastseen = set([])
    gwv_last_reward = 0

    height, width = get_height_width(world)

    if env is not None:
        """if action == pick:
            minigrid_digest(env.step(action_pick))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_pick))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_pick))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_pick))
            minigrid_digest(env.step(action_right))
        if action == drop:
            minigrid_digest(env.step(action_drop))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_drop))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_drop))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_drop))
            minigrid_digest(env.step(action_right))
        if action == toggle:
            minigrid_digest(env.step(action_toggle))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_toggle))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_toggle))
            minigrid_digest(env.step(action_right))
            minigrid_digest(env.step(action_toggle))
            minigrid_digest(env.step(action_right))"""

        if action == drop:  # drop below always
            if gwv_direction == dir_down:
                help(action);
                minigrid_digest(env.step(action_drop))
            elif gwv_direction == dir_left:
                minigrid_digest(env.step(action_left))
                help(action);
                minigrid_digest(env.step(action_drop))
            elif gwv_direction == dir_right:
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_drop))
            elif gwv_direction == dir_up:
                minigrid_digest(env.step(action_right))
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_drop))
        if action == left:
            if gwv_direction == dir_left:
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_down:
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_up:
                minigrid_digest(env.step(action_left))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_right:
                minigrid_digest(env.step(action_right))
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            """if lastActionIsPick:
                minigrid_digest(env.step(action_pick))
            if lastActionIsDrop:
                minigrid_digest(env.step(action_drop))
            if lastActionIsToggle:
                minigrid_digest(env.step(action_toggle))"""
        if action == right:
            if gwv_direction == dir_right:
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_down:
                minigrid_digest(env.step(action_left))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_up:
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_left:
                minigrid_digest(env.step(action_right))
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            """if lastActionIsPick:
                minigrid_digest(env.step(action_pick))
            if lastActionIsDrop:
                minigrid_digest(env.step(action_drop))
            if lastActionIsToggle:
                minigrid_digest(env.step(action_toggle))"""
        if action == up:
            if gwv_direction == dir_up:
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_right:
                minigrid_digest(env.step(action_left))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_left:
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_down:
                minigrid_digest(env.step(action_right))
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            """if lastActionIsPick:
                minigrid_digest(env.step(action_pick))
            if lastActionIsDrop:
                minigrid_digest(env.step(action_drop))
            if lastActionIsToggle:
                minigrid_digest(env.step(action_toggle))"""
        if action == down:
            if gwv_direction == dir_down:
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_right:
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_left:
                minigrid_digest(env.step(action_left))
                help(action);
                minigrid_digest(env.step(action_forward))
            if gwv_direction == dir_up:
                minigrid_digest(env.step(action_right))
                minigrid_digest(env.step(action_right))
                help(action);
                minigrid_digest(env.step(action_forward))
            """if lastActionIsPick:
                minigrid_digest(env.step(action_pick))
            if lastActionIsDrop:
                minigrid_digest(env.step(action_drop))
            if lastActionIsToggle:
                minigrid_digest(env.step(action_toggle))"""
        # help(action)
        newloc = env.agent_pos  # a breakpoint placed here is never triggered from any of my flows.
        oldworld = deepcopy(world)

        M = {(1, 0): FREE, (2, 0): WALL, (4, 0): FREE, (4, 1): FREE, (4, 2): DOOR, (5, 0): KEY, (6, 0): BALL,
             (7, 0): TABLE, (8, 0): GOAL, (9, 0): SHOCK}
        for i in range(7):
            for j in reversed(range(7)):
                if lastimage[i, j][0] == 0:
                    break
                if gwv_direction == dir_right:
                    X = newloc[0] + (7 - (j + 1))
                    Y = newloc[1] + i - 3
                    gwv_lastseen.add((Y, X))
                    # print(lastimage); exit(0)
                    V = (lastimage[i, j][0], lastimage[i, j][2])
                    if V[0] != 0 and Y >= 0 and X >= 0 and Y < height and X < width:
                        # print("!!!", (X,Y), (i,j), V)
                        world[BOARD][Y][X] = M[V]
                if gwv_direction == dir_left:
                    X = newloc[0] - (7 - (j + 1))
                    Y = newloc[1] - i + 3
                    # print(lastimage); exit(0)
                    gwv_lastseen.add((Y, X))
                    V = (lastimage[i, j][0], lastimage[i, j][2])
                    if V[0] != 0 and Y >= 0 and X >= 0 and Y < height and X < width:
                        # print("!!!", (X,Y), (i,j), V)
                        world[BOARD][Y][X] = M[V]

                if gwv_direction == dir_up:
                    Y = newloc[1] - (7 - (j + 1))
                    X = newloc[0] + i - 3
                    # print(lastimage); exit(0)
                    gwv_lastseen.add((Y, X))
                    V = (lastimage[i, j][0], lastimage[i, j][2])
                    if V[0] != 0 and Y >= 0 and X >= 0 and Y < height and X < width:
                        # print("!!!", (X,Y), (i,j), V)
                        world[BOARD][Y][X] = M[V]
                if gwv_direction == dir_down:
                    Y = newloc[1] + (7 - (j + 1))
                    X = newloc[0] - i + 3
                    # print(lastimage); exit(0)
                    gwv_lastseen.add((Y, X))
                    V = (lastimage[i, j][0], lastimage[i, j][2])
                    if V[0] != 0 and Y >= 0 and X >= 0 and Y < height and X < width:
                        # print("!!!", (X,Y), (i,j), V)
                        world[BOARD][Y][X] = M[V]

        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set
        world[BOARD][newloc[1]][newloc[0]] = ROBOT
        local_location = newloc
        i_inventory = 3
        j_inventory = 6
        V_inventory = lastimage[i_inventory, j_inventory][0]
        if (cntEntry(oldworld, TABLE) > cntEntry(world, TABLE)) or \
                (cntEntry(oldworld, GOAL) > cntEntry(world, GOAL)) or \
                (cntEntry(oldworld, SHOCK) > cntEntry(world, SHOCK)) or gwv_last_reward != 0:
            gwv_episode_id += 1
            if gwv_last_reward > 0 or (cntEntry(oldworld, TABLE) > cntEntry(world, TABLE)) or (
                    cntEntry(oldworld, GOAL) > cntEntry(world, GOAL)):
                gwv_last_reward = 1  # minigrid is not giving so we provide own reward
            else:
                gwv_last_reward = -1
            gwv_accumulated_score += gwv_last_reward
            world[VALUES] = (gwv_accumulated_score, V_inventory)
            gwv_has_reset = 2  # ugly double reset hack is needed
            gwv_reset_score = world[VALUES][0]
            minigrid_digest(env.reset())
            if run_id != -1:
                reward = 1 - 0.9 * (gwv_step / max_steps)
                if gwv_episode_id < episodes:
                    with open("run_world" + str(_challenge_input) + "_" + str(run_id) + ".run", "a") as f:
                        f.write(f"{gwv_episode_id} {reward}\n")
                if episodes != -1 and gwv_episode_id >= episodes:
                    exit(0)
                gwv_step = 0
        else:
            gwv_accumulated_score += gwv_last_reward
            world[VALUES] = (gwv_accumulated_score, V_inventory)
        return local_location, [world[BOARD], world[VALUES], world[TIMES]]
    if _slippery and random.random() > 0.9:  # agent still believes it did the proper action
        action = random.choice(gwv_full_action_list)  # but the world is slippery!
    newloc = action(local_location)  # the bounding is checked just below

    if newloc[0] < 0:
        newloc = (0, newloc[1])
    if newloc[1] < 0:
        newloc = (newloc[0], 0)
    if newloc[0] >= width:
        newloc = (width - 1, newloc[1])
    if newloc[1] >= height:
        newloc = (newloc[0], height - 1)

    oldworld = deepcopy(world)
    # ROBOT MOVEMENT ON FREE SPACE
    movedAgent = False
    if oldworld[BOARD][newloc[1]][newloc[
        0]] in get_traversable_board_values():  # TODO the list of objects the agent can move onto should be learned
        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set  # TODO after the agent moved off a spot, replace it with the last observed in that location? or free, and let outside model correct it?
        world[BOARD][newloc[1]][newloc[0]] = ROBOT
        movedAgent = local_location != newloc
        local_location = newloc
    oldoldworld = deepcopy(oldworld)
    oldworld = deepcopy(world)

    annomate_arrows = random.random() < annomate_arrows_prob

    for y in range(height):
        for x in range(width):
            if oldworld[BOARD][y][x] == BALL and world[BOARD][y][x - 1] in gwv_traversable_board_values and \
                    oldoldworld[BOARD][y][
                        x - 1] in gwv_traversable_board_values:
                world[BOARD][y][x - 1] = BALL
                world[BOARD][y][x] = gwv_traversable_board_values[0]  # set to first free value from transversable set
            if oldworld[BOARD][y][x] == BALL and oldworld[BOARD][y][x - 1] == WALL:
                world[BOARD][y][x] = gwv_traversable_board_values[0]  # set to first free value from transversable set
                world[BOARD][random.choice(range(1, height - 1))][width - 1] = BALL
            if oldworld[BOARD][y][x] == BALL and oldoldworld[BOARD][y][x - 1] == ROBOT:
                world[BOARD][y][x] = gwv_traversable_board_values[0]  # set to first free value from transversable set
                world[BOARD][random.choice(range(1, height - 1))][width - 1] = BALL
                if not movedAgent:
                    world[VALUES] = tuple(
                        [world[VALUES][0] + 1] + list(world[VALUES][1:]))  # the first value +1 and the rest stays

            if annomate_arrows and oldworld[BOARD][y][x] == ARROW_DOWN and oldworld[BOARD][y + 1][
                x] in gwv_traversable_board_values:
                world[BOARD][y + 1][x] = ARROW_DOWN
                world[BOARD][y][x] = gwv_traversable_board_values[0]  # set to first free value from transversable set
            if annomate_arrows and oldworld[BOARD][y][x] == ARROW_UP and oldworld[BOARD][y - 1][
                x] in gwv_traversable_board_values:
                world[BOARD][y - 1][x] = ARROW_UP
                world[BOARD][y][x] = gwv_traversable_board_values[0]  # set to first free value from transversable set
            if annomate_arrows and oldworld[BOARD][y][x] == ARROW_DOWN and oldworld[BOARD][y + 1][x] in [WALL, ROBOT,
                                                                                                         FOOD]:
                world[BOARD][y][x] = ARROW_UP
            if annomate_arrows and oldworld[BOARD][y][x] == ARROW_UP and oldworld[BOARD][y - 1][x] in [WALL, ROBOT,
                                                                                                       FOOD]:
                world[BOARD][y][x] = ARROW_DOWN
            if oldworld[BOARD][y][x] == CUP and oldworld[BOARD][y + 1][x] == TABLE:
                if not _isWorld9:
                    world[BOARD][y][x] = gwv_traversable_board_values[
                        0]  # set to first free value from transversable set
                    while True:
                        xr, yr = (random.randint(0, width - 1), random.randint(0, height - 1))
                        if oldworld[BOARD][yr][xr] in gwv_traversable_board_values:
                            world[BOARD][yr][xr] = CUP
                            break
    # CUP
    if world[BOARD][newloc[1]][newloc[0]] == CUP:  # an object the system could shift around
        world[BOARD][local_location[1]][local_location[0]] = CUP
        local_location = newloc
        world[BOARD][local_location[1]][local_location[0]] = ROBOT
    # KEY
    if world[BOARD][newloc[1]][newloc[0]] == KEY:
        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set
        local_location = newloc
        world[BOARD][local_location[1]][local_location[0]] = ROBOT
        world[VALUES] = tuple([world[VALUES][0]] + [world[VALUES][1] + 1] + list(
            world[VALUES][2:]))  # the second value +1 and the rest stays. (we picked up a key)
    # DOOR
    if world[BOARD][newloc[1]][newloc[0]] == DOOR and world[VALUES][1] > 0:
        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set
        local_location = newloc
        world[BOARD][local_location[1]][local_location[0]] = ROBOT
        world[VALUES] = tuple([world[VALUES][0]] + [world[VALUES][1] - 1] + list(
            world[VALUES][2:]))  # the second value -1 and the rest stays (we used the key on the door)
    # BATTERY
    if world[BOARD][newloc[1]][newloc[0]] == BATTERY:
        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set
        local_location = newloc
        world[BOARD][local_location[1]][local_location[0]] = ROBOT
        world[VALUES] = tuple([world[VALUES][0] + 1] + list(world[VALUES][1:]))  # the first value +1 and the rest stays
    # SHOCK
    if world[BOARD][newloc[1]][newloc[0]] == SHOCK:
        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set
        local_location = newloc
        world[BOARD][local_location[1]][local_location[0]] = ROBOT
        world[VALUES] = tuple([world[VALUES][0] - 1] + list(world[VALUES][1:]))  # the first value -1 and the rest stays
    # FOOD
    if world[BOARD][newloc[1]][newloc[0]] == FOOD:
        world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
            0]  # set to first free value from transversable set
        local_location = newloc
        world[BOARD][local_location[1]][local_location[0]] = ROBOT
        world[VALUES] = tuple([world[VALUES][0] + 1] + list(world[VALUES][1:]))  # the first value +1 and the rest stays

        new_food_placed = False
        for poss_xy_location in [(8, 1), (8, 5), (9, 5), (7, 5)]:
            x, y = poss_xy_location
            if world[BOARD][y][x] in gwv_traversable_board_values:
                world[BOARD][y][x] = FOOD
                new_food_placed = True
                break
        if not new_food_placed:
            while True:
                x, y = (random.randint(0, width - 1), random.randint(0, height - 1))
                if world[BOARD][y][x] in gwv_traversable_board_values:
                    world[BOARD][y][x] = FOOD
                    break

    # EGG
    if world[BOARD][newloc[1]][newloc[0]] == EGG and world[VALUES][1] == 0:  # can only carry 1
        world[BOARD][newloc[1]][newloc[0]] = EGGPLACE  # EGGPLACE - where an egg was.
        world[VALUES] = tuple([world[VALUES][0]] + [world[VALUES][1] + 1] + list(world[VALUES][2:]))
    elif world[BOARD][newloc[1]][newloc[0]] == EGGPLACE and world[VALUES][1] > 0:
        world[BOARD][newloc[1]][newloc[0]] = EGG
        world[VALUES] = tuple([world[VALUES][0]] + [world[VALUES][1] - 1] + list(world[VALUES][2:]))
    elif world[BOARD][newloc[1]][newloc[0]] == CHICKEN and world[VALUES][1] > 0:
        world[VALUES] = tuple([world[VALUES][0]] + [world[VALUES][1] - 1] + list(world[VALUES][2:]))
        world[VALUES] = tuple([world[VALUES][0] + 1] + list(world[VALUES][1:]))  # 1 food
    # Football
    crateloc = action(newloc)
    if crateloc[1] < height and crateloc[0] < width and crateloc[1] >= 0 and crateloc[0] >= 0:
        if world[BOARD][crateloc[1]][crateloc[0]] in gwv_traversable_board_values:
            if world[BOARD][newloc[1]][newloc[0]] == SBALL:
                world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
                    0]  # set to first free value from transversable set
                local_location = newloc
                world[BOARD][local_location[1]][local_location[0]] = ROBOT
                world[BOARD][crateloc[1]][crateloc[0]] = SBALL
        if world[BOARD][crateloc[1]][crateloc[0]] == GOAL and world[BOARD][newloc[1]][newloc[0]] == SBALL:
            world[BOARD][local_location[1]][local_location[0]] = gwv_traversable_board_values[
                0]  # set to first free value from transversable set
            local_location = newloc
            world[BOARD][newloc[1]][newloc[0]] = ROBOT
            world[VALUES] = tuple(
                [world[VALUES][0] + 1] + list(world[VALUES][1:]))  # the first value +1 and the rest stays
            while True:
                xr, yr = (random.randint(0, width - 1), random.randint(0, height - 1))
                if oldworld[BOARD][yr][
                    xr] in gwv_traversable_board_values and xr > 1 and yr > 1 and xr < width - 2 and yr < height - 2 and \
                        oldworld[BOARD][yr + 1][xr] in gwv_traversable_board_values and oldworld[BOARD][yr - 1][
                    xr] in gwv_traversable_board_values and \
                        oldworld[BOARD][yr][xr + 1] in gwv_traversable_board_values and oldworld[BOARD][yr][
                    xr - 1] in gwv_traversable_board_values:
                    world[BOARD][yr][xr] = SBALL
                    break

    if external_reward_for_last_action is not None:
        world[VALUES] = tuple(
            [world[VALUES][0] + external_reward_for_last_action] + list(
                world[VALUES][1:]))  # the first value += external_reward_for_last_action and the rest stays

    if World_GetObjective() is not None:
        if World_GetObjective()(world):
            l = list(world[VALUES])
            world[VALUES] = tuple([l[0] + 1] + l[1:])

    return local_location, [world[BOARD], world[VALUES], world[TIMES]]


def World_Criteria(world):
    return World_objective(world)


def World_SetObjective(func):
    global World_objective
    World_objective = func


def World_GetObjective():
    return World_objective


# Print the world into the terminal
def World_Print(world):
    for line in world[BOARD]:
        print("".join(line))


def Multiple_World_Print(list_of_worlds, pad_length=30):
    """
    Print a number of worlds in coloured text, left to right across the screen
    @param list_of_worlds: dict e.g. [{"World":list[list[char]], "Caption":str, "Color":color code}]
    @return:
    """
    max_lines = max([len(record["World"][BOARD]) for record in list_of_worlds])

    line = ""
    for world_index in range(len(list_of_worlds)):
        line += list_of_worlds[world_index]["Caption"].ljust(pad_length, " ")

    print(line)

    for line_num in range(max_lines):
        line = ""
        for world_index in range(len(list_of_worlds)):
            pass
            # Set the color
            if "Color" in list_of_worlds[world_index]:
                line += list_of_worlds[world_index]["Color"]
            map_line = "".join(list_of_worlds[world_index]["World"][BOARD][line_num])
            line += map_line
            line += nace.color_codes.color_code_black_on_white
            line += map_line.ljust(pad_length, " ")[len(map_line):]
        print(line)
    pass


def World_FieldOfView(Time, local_xy_location, observed_world, external_world):
    """
    The limited field of view the agent can observe dependent on view distance (partial observability)
    Does this by merging parts of the full visibility world with the last known partially observed world.
    State is held in the observed world. New visible portions of 'world' are copied into observed world.

    This mutates the observed_world passed in (updated time) and returns it, but assignment of the return value to the same variable
    can overwrite these changes.

    @param Time:
    @param local_xy_location:
    @param observed_world:
    @param external_world:
    @return:
    """
    global gwv_has_reset

    height, width = (len(external_world[BOARD]), len(external_world[BOARD][0]))

    for Y in range(height):
        for X in range(width):
            if observed_world[TIMES][Y][X] == Time:
                observed_world[TIMES][Y][
                    X] = Time - 1  # WHY can this ever happen??? DEBUG! it can happen if this routine is called twice in a row.
    if env is None:  # usual case. This will 'move' the robot in the process of freshening the observed world.
        for y in range(VIEWDISTY * 2 + 1):
            for x in range(VIEWDISTX * 2 + 1):
                Y = local_xy_location[1] + y - VIEWDISTY
                X = local_xy_location[0] + x - VIEWDISTX
                if Y >= 0 and Y < height and \
                        X >= 0 and X < width:
                    observed_world[BOARD][Y][X] = external_world[BOARD][Y][X]
                    observed_world[TIMES][Y][X] = Time
    else:
        if gwv_has_reset > 0:
            external_world[VALUES] = (gwv_reset_score, 0)
            for y in range(height):
                for x in range(width):
                    observed_world[BOARD][y][x] = '.'
                    external_world[BOARD][y][x] = gwv_traversable_board_values[
                        0]  # set to first free value from transversable set
                    observed_world[TIMES][y][x] = Time - 1000
            observed_world[VALUES] = (gwv_reset_score, 0)
            external_world[VALUES] = (gwv_reset_score, 0)
            gwv_has_reset -= 1
            # PRINT("RESET!!!!!!!!!!!!")
        for Y in range(height):
            for X in range(width):
                if observed_world[BOARD][Y][X] == ROBOT:
                    observed_world[BOARD][Y][X] = gwv_traversable_board_values[
                        0]  # set to first free value from transversable set
        for Y in range(height):
            for X in range(width):
                if (Y, X) in gwv_lastseen:
                    if Y >= 0 and Y < height and \
                            X >= 0 and X < width:
                        observed_world[BOARD][Y][X] = external_world[BOARD][Y][X]
                        observed_world[TIMES][Y][X] = Time
    observed_world[VALUES] = deepcopy(external_world[VALUES])  # copy the values across (NOT the times or board)


def World_AsTuple(worldpart):
    """
    @param worldpart:
    @return: The world component represented as an immutable tuple
    """
    return tuple(World_AsTuple(i) if isinstance(i, list) else i for i in worldpart)


def World_Num5():
    return _isWorld5


# The actions the agent can take dependent on the chosen world:
gwv_full_action_list = [left, right, up, down]  # was called actions


def get_full_action_list():
    """
    @return: a list of all the known/valid actions for the world
    """
    global gwv_full_action_list

    # if (gwv_full_action_list[0] != left or
    #         gwv_full_action_list[1] != right or
    #         gwv_full_action_list[2] != up or
    #         gwv_full_action_list[3] != down
    # ):
    #     print("WARN: Actions are stored in a different order")

    return copy.deepcopy(gwv_full_action_list)


def add_to_full_action_list(new_action):
    """
    Adds an action to this list of existing actions if not already part of the list.
    @param new_action:
    @return: nothing
    """
    global gwv_full_action_list
    if new_action not in gwv_full_action_list:
        gwv_full_action_list.append(new_action)


def set_full_action_list(action_list):
    """
    sets in the list of passed in actions as the possible action that can be taken. Conserves order.
    @param new_action:
    @return: nothing
    """
    global gwv_full_action_list
    gwv_full_action_list.clear()
    gwv_full_action_list.extend(action_list)


# The board values that the agent can move onto. Use getters and setters for access
gwv_traversable_board_values = list(set([FREE]))


def add_traversable_board_value(value: str):
    global gwv_traversable_board_values
    gwv_traversable_board_values.add(value)


def set_traversable_board_value(value: str):
    global gwv_traversable_board_values
    gwv_traversable_board_values = list(set([value]))


def get_traversable_board_values():
    global gwv_traversable_board_values
    return deepcopy(gwv_traversable_board_values)



def _act(
        xy_loc, fully_observed_pre_action_world, action, inject_key, external_reward_for_last_action):
    # move the agent on the map, and upload location
    xy_loc, fully_observed_post_action_world = nace.world_module.World_Move(xy_loc,
                                                                       deepcopy(fully_observed_pre_action_world),
                                                                       action,
                                                                       None if external_reward_for_last_action == 0 else external_reward_for_last_action)
    return xy_loc, fully_observed_post_action_world, "na"



import copy


class Agent():
    """
    Holds the state of the agent, including location, and 'internal' arbitrary values (num keys held)
    """

    def __init__(self, xy_loc, score, values_exc_score):
        self.xy_loc = xy_loc
        self.values = [score] + list(values_exc_score)

    def set_xy_loc(self, xy_loc):
        self.xy_loc = xy_loc

    def get_xy_loc(self):
        return copy.deepcopy(self.xy_loc)

    def set_score(self, score):
        self.values[0] = score

    def get_score(self):
        return self.values[0]

    def get_values_exc_score(self):
        # adding anything here may explode the size of the planning space.
        return copy.deepcopy(self.values[1:])

    def get_values_inc_score(self):
        # extending the length of the values may explode the size of the planning space.
        return copy.deepcopy(self.values)

    def set_values_inc_score(self, values):
        self.values = list(copy.deepcopy(values))

    def increment_values_including_score(self, values):
        for i in range(len(values)):
            if i < len(self.values):
                if values[i] > 0:
                    pass
                self.values[i] += values[i]



import copy

import numpy as np

import nace.color_codes

UNOBSERVED_BOARD_VALUE = '.'


class NPWorld():

    def __init__(self, with_observed_time, name, width=None, height=None, initial_value=ord(UNOBSERVED_BOARD_VALUE),
                 view_dist_x=3, view_dist_y=2):
        self.name = name
        if width is None or height is None:
            self.board = np.zeros((0, 0), dtype=int)
        else:
            self.board = np.zeros((width, height), dtype=int)
            self.board[:] = initial_value

        self.initial_value = initial_value
        self.with_observed_time = with_observed_time

        self.times = np.zeros((0, 0), dtype=np.float16)  # this seems expensive in space, a float16 for each cell. hmmm
        self.times[:] = float('-inf')
        self.agent_location_column = -1
        self.agent_location_row = -1
        self.view_dist_x = view_dist_x
        self.view_dist_y = view_dist_y
        self.debug_board = ""

    def get_char_at(self, row, col):
        return chr(self.board[row, col])

    def set_char_at(self, row, col, c, location_indicator_char='x'):
        self.board[row, col] = ord(c)
        if c == location_indicator_char:
            self.agent_location_column = col
            self.agent_location_row = row

    def get_time_at(self, row, col):
        return self.times[row, col]

    def get_newest_time(self):
        """
        @return: the must recent, up-to-date time stored in the world.
        """
        if self.with_observed_time:
            if self.times.shape[0] > 0:
                return self.times.max()
        return 0

    def get_board_line(self, row_number, color):
        if self.with_observed_time and self.times.size > 0:
            oldest_time = self.times.min().item()
        else:
            oldest_time = float('-inf')
        line = ""
        color_count = 0
        if row_number < self.board.shape[0]:
            for col in range(self.board.shape[1]):
                if self.with_observed_time and self.times[row_number][col] == oldest_time:
                    line += nace.color_codes.color_code_black_on_white
                    color_count += 1
                line += chr(self.board[row_number, col])

                if self.with_observed_time and self.times[row_number][col] == oldest_time:
                    if color != None:
                        line += color
        return line, self.board.shape[1]

    def get_in_nace_board_format(self):
        board = []
        for row in range(self.board.shape[0]):
            line = []
            for col in range(self.board.shape[1]):
                line.append(chr(self.board[row, col]))
            board.append(line)
        return board

    def _reset_non_board(self):
        # set the times and other values to the default
        if self.with_observed_time:
            self.times = np.zeros(self.board.shape, dtype=np.float16)  # this seems expensive in space. hmmm
            self.times[:] = float('-inf')
        self.agent_location_column = -1
        self.agent_location_row = -1

    def update_world_from_ground_truth_gymnasium_format(
            self,
            gym_world: str,
            strip_blanks=True,
            location_indicator_char='x',
            add_surrounding_walls=True,
            wall_code='W',
            time_counter: float = float('-inf')):
        """

        Convert a world in gym format, into a numpy one (via the nace format)

        @param gym_world: str , each \n indicates a new row, white space is stripped. agent location indicated by  location_indicator_char
        @param strip_blanks:
        @param location_indicator_char:  char representing the agent
        @param last_world: if not none, state and times seen copied from this world
        @param add_walls_right_and_bottom:
        @param wall_code:
        @return: the location of last agent seen
        """
        list_of_list_of_str = []  # list of rows. Each row is a list of char.
        for row_number, row_str in enumerate(gym_world.split("\n")):
            if len(row_str.strip()) > 0:
                if strip_blanks:
                    list_of_list_of_str.append([c for c in list(row_str) if c != " "])
                else:
                    list_of_list_of_str.append(list(row_str))

        if add_surrounding_walls:
            for row in list_of_list_of_str:
                row.append(wall_code)
                row.insert(0, wall_code)
            extra_line = [wall_code for _ in row]
            list_of_list_of_str.append(extra_line)
            list_of_list_of_str.insert(0, extra_line)

        # check the sizes are correct, and see where location marker is
        for row_number in range(len(list_of_list_of_str)):
            if row_number > 0:
                assert len(list_of_list_of_str[0]) == len(list_of_list_of_str[row_number])

        # the next line returns the xy location of last agent seen, modified count, and pre-action world
        return self.update_world_from_ground_truth_nace_format(list_of_list_of_str, time_counter=time_counter,
                                                               location_indicator_char=location_indicator_char)

    def _set_board_size(self, height, width):
        prior_board_shape = self.board.shape
        size_changed = False
        if prior_board_shape == (0, 0):
            # we need to reset all other values
            self.board = np.zeros((height, width), dtype=int)
            self.board[:] = self.initial_value
            self._reset_non_board()
            size_changed = True

        return prior_board_shape, size_changed

    def update_world_from_ground_truth_nace_format(
            self,
            nace_format_world_list_of_list_of_str,  # just the board
            location_indicator_char='x',
            time_counter=float('-inf'),
            times_list_of_str=None
    ):
        """
        Update the board with the values passed in when the passed in value is not the unobserved marker
        The time the update happened is written to the times array.

        @param nace_format_world_list_of_list_of_str:
        @param location_indicator_char:
        @param unobserved_indicator_char:
        @param time_counter:
        @return:
        """

        prior_board_shape, size_changed = self._set_board_size(
            height=len(nace_format_world_list_of_list_of_str),
            width=len(nace_format_world_list_of_list_of_str[0]))

        pre_update_world = copy.deepcopy(self)

        location_column = None
        location_row = None
        modified_count = 0

        # find the agents location (assumes only one agent)
        for row_number in range(len(nace_format_world_list_of_list_of_str)):
            for col_number in range(len(nace_format_world_list_of_list_of_str[row_number])):
                c = nace_format_world_list_of_list_of_str[row_number][col_number]
                if c == location_indicator_char:
                    location_column = col_number
                    location_row = row_number

        # iterate only over the observable rows and columns,
        # updating board and times if they are inside the view distance
        for y_delta in range(self.view_dist_y * 2 + 1):
            row_number = location_row + y_delta - self.view_dist_y
            if row_number >= 0 and row_number < self.board.shape[0]:
                for x_delta in range(self.view_dist_x * 2 + 1):
                    col_number = location_column + x_delta - self.view_dist_x
                    if col_number >= 0 and col_number < self.board.shape[1]:
                        c = nace_format_world_list_of_list_of_str[row_number][col_number]
                        if ord(c) != self.initial_value:
                            if self.board[row_number][col_number] != ord(c):
                                self.board[row_number][col_number] = ord(c)
                                modified_count += 1
                            if self.with_observed_time:
                                self.times[row_number][col_number] = float(time_counter)
                        else:  # unobserved
                            pass  # leave the observed time, and board value as they are.

        if self.with_observed_time:
            if times_list_of_str is not None:
                # set the times even though assumed to be only for testing
                print("WARN: setting observation times. This code path should only be triggered in unit tests")
                for row_number in range(len(times_list_of_str)):
                    times = times_list_of_str[row_number].split(",")
                    for col_number in range(len(times)):
                        s = times[col_number].strip()
                        if s != '':
                            self.times[row_number][col_number] = float(s)

        if prior_board_shape != (0, 0):
            if prior_board_shape != self.board.shape:
                raise Exception("Changing board size/shape in flight not yet supported.")
        if location_column is not None and location_row is not None:
            self.agent_location_column = location_column
            self.agent_location_row = location_row

        self.debug_board = self.board_as_string()

        return (location_column, location_row), modified_count, pre_update_world

    def update_world_from_ground_truth(self, time_counter, external_ground_truth_world_model, xy_locations,
                                       location_indicator_char='x'):

        # todo should we pass xy_loc in here? or calculate it? or use the stored version?
        height = external_ground_truth_world_model.board.shape[0]
        width = external_ground_truth_world_model.board.shape[1]

        self._set_board_size(
            height=height,
            width=width)

        pre_action_internal_world = copy.deepcopy(self)

        # for Y in range(height): this was present in the old code
        #     for X in range(width):
        #         if observed_world[TIMES][Y][X] == Time:
        #             observed_world[TIMES][Y][
        #                 X] = Time - 1  # WHY can this ever happen??? DEBUG! it can happen if this routine is called twice in a row.
        modified_count = 0
        for y_delta in range(self.view_dist_y * 2 + 1):
            for x_delta in range(self.view_dist_x * 2 + 1):
                for xy_loc in xy_locations:
                    y = xy_loc[1] + y_delta - self.view_dist_y
                    x = xy_loc[0] + x_delta - self.view_dist_x
                    if y >= 0 and y < height and \
                            x >= 0 and x < width:
                        if external_ground_truth_world_model.board[y][x] != ord(UNOBSERVED_BOARD_VALUE):
                            if self.board[y][x] != external_ground_truth_world_model.board[y][x]:
                                self.board[y][x] = external_ground_truth_world_model.board[y][x]
                                modified_count += 1

                            if self.with_observed_time:
                                self.times[y][x] = time_counter
                        if chr(self.board[y][x]) == location_indicator_char:
                            self.agent_location_column = x
                            self.agent_location_row = y

        if modified_count == 0:
            pass
        self.debug_board = self.board_as_string()
        return modified_count, pre_action_internal_world

    def get_agent_xy_location(self):  # note there may be more than 1 agent at some stage.
        return (self.agent_location_column, self.agent_location_row)

    def get_list_of_differences(self, world2):
        differences = self.board != world2.board
        (row_indexes, column_indexes) = np.nonzero(
            differences)  # return 2 vectors, with all the y values in one, and x in the other
        return row_indexes, column_indexes

    def get_board_hashcode(self):
        """
        Note this may return different hashes on different runs. To Be Checked
        @return:
        """
        return hash(bytes(self.board.data))

    def get_height_width(self):
        return self.board.shape[0], self.board.shape[1]

    def get_board_char_counts(self):
        result = {}
        unique_values = np.unique(self.board)
        for v in unique_values:
            s = np.sum(self.board == v)
            result[chr(v)] = s.item()
        return result

    def multiworld_print(self, list_of_records, pad_length=30):
        """
        Print a number of worlds in coloured text, left to right across the screen
        @param list_of_worlds: dict e.g. [{"World":list[list[char]], "Caption":str, "Color":color code}]
        @return:
        """
        height = [record["World"].get_height_width()[0] for record in list_of_records]
        max_lines = max(height)

        for caption_line in range(3):
            line = ""
            for world_index in range(len(list_of_records)):
                caption = ""
                if "Caption" in list_of_records[world_index]:
                    caption_list = list_of_records[world_index]["Caption"].split("\n")
                    caption_list.extend(["", "", ""])
                    caption = caption_list[caption_line]
                line += caption.ljust(pad_length, " ")
            print(line)

        for line_num in range(max_lines):
            line = ""
            for world_index in range(len(list_of_records)):
                pass
                # Set the color
                color = None
                if "Color" in list_of_records[world_index]:
                    color = list_of_records[world_index]["Color"]
                    line += color
                map_line, row_length = list_of_records[world_index]["World"].get_board_line(line_num, color)
                line += map_line
                line += nace.color_codes.color_code_black_on_white

                padding = "".join([" "] * (pad_length - row_length))
                line += padding
                # line += map_line.ljust(pad_length, " ")[len(map_line):]
            print(line)
        pass

    def board_as_string(self):
        result = ""
        for row_number in range(self.board.shape[0]):
            for col_number in range(self.board.shape[1]):
                result += chr(self.board[row_number, col_number])
            result += "\n"
        return result

    def to_string_list(self):
        board_lines = []
        time_lines = []
        for row_number in range(self.board.shape[0]):
            board_line = ""
            time_line = ""
            for col_number in range(self.board.shape[1]):
                c = chr(self.board[row_number][col_number])
                board_line += c
                t = self.times[row_number][col_number]
                if t == float("-inf"):
                    time_line += "-inf,"
                else:
                    time_line += str(t) + ","
            board_lines.append(board_line)
            time_lines.append(time_line)
        return ([board_lines, (), time_lines])

    @staticmethod
    def from_string_list(str_list: str, location_indicator_char="x", unobserved_indicator_char='.', view_dist_x=3,
                         view_dist_y=2):
        """

        @param str_list: tuple of (board, values, times) where board is in the format:
                     ['oooooooooooo',
                     'of  o      o',
                     'o          o',
                     'o   oooooooo',
                     'o       u  o',
                     'o      x   o',
                     'oooooooooooo'],
        @param location_indicator_char:
        @param unobserved_indicator_char:
        @param view_dist_x:
        @param view_dist_y:
        @return:
        """
        world = NPWorld(with_observed_time=True, name="from_string()", initial_value=ord(unobserved_indicator_char),
                        view_dist_x=view_dist_x, view_dist_y=view_dist_y)
        board_list_of_str = str_list[0]
        times_list_of_str = str_list[2]
        world.update_world_from_ground_truth_nace_format(
            board_list_of_str,
            time_counter=float('-inf'),  # -inf == never seen
            location_indicator_char=location_indicator_char,
            times_list_of_str=times_list_of_str)

        return world

    def from_gymnasium(gym_world: str,
                       strip_blanks=True,
                       location_indicator_char='x',
                       state_source=None,
                       add_surrounding_walls=True,
                       wall_code='W',
                       observed_at=0):
        """

        @param gym_world: str , each \n indicates a new row, white space is stripped. agent location indicated by  location_indicator_char
        @param strip_blanks:
        @param location_indicator_char:  char representing the agent
        @param last_world: if not none, state and times seen copied from this world
        @param add_walls_right_and_bottom:
        @param wall_code:
        @return:
        """
        board = []  # list of rows. Each row is a list of char.
        for row_number, row_str in enumerate(gym_world.split("\n")):
            if len(row_str.strip()) > 0:
                if strip_blanks:
                    board.append([c for c in list(row_str) if c != " "])
                else:
                    board.append(list(row_str))
        if add_surrounding_walls:
            for row in board:
                row.append(wall_code)
                row.insert(0, wall_code)
            extra_line = [wall_code for _ in row]
            board.append(extra_line)
            board.insert(0, extra_line)

        BOARD, VALUES, TIMES = (0, 1, 2)

        if state_source is None:
            world_state = tuple([0, 0])
        else:
            world_state = copy.deepcopy(state_source[VALUES])

        height, width = (len(board), len(board[0]))
        world_observed_at = [[observed_at for i in range(width)] for j in range(height)]
        str_list = [board, world_state, world_observed_at]

        # xy_location = extract_agent_location_from_board(board, location_indicator_char=location_indicator_char)
        xy_location = () # TODO

        nace_world = NPWorld.from_string_list(str_list, location_indicator_char=location_indicator_char,)

        return nace_world, xy_location  # NOTE location is xy format




color_code_black_on_white = f"\033[0m"
color_code_white_on_red = f"\033[97;41m"
color_code_white_on_black = f"\033[97;40m"
color_code_white_on_blue = f"\033[97;44m"

color_code_black_on_white_v2 = "\033[1;1H\033[2J"



from setuptools import setup, find_packages
import base64

install_requires = ['numba>=0.60.0',
                    # "gymnasium[classic-control]"
                    ]


from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README_pypi.md").read_text()





setup_kwargs = {
    'name': 'nace',
    'version': '0.0.17',
    'description': 'A re-implementation of NACE, as a pypi package, with a cleaner more general interface.', # overall description
    'long_description': long_description,
    'long_description_content_type':'text/markdown',
    'author': 'ucabdv1',
    'author_email': 'ucabdv1@ucl.ac.uk',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': find_packages(),
    'install_requires': install_requires,
    'python_requires': '>=3.10',
    'classifiers':[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            ]
}


setup(**setup_kwargs)


from hypothesis import Hypothesis_TruthValue
from prettyprint import Prettyprint_rule
from world_module import right

if __name__ == "__main__":
    rule_evidence = {((right, (0,), (0, 0, 'x'), (0, 1, 'o')), (0, 0, ' ', (-1, 0))): (1, 0)}
    rule = ((right, (0,), (0, 0, 'x'), (0, 1, 'o')), (0, 0, ' ', (-1, 0)))

    Prettyprint_rule(rule_evidence, Hypothesis_TruthValue, rule)



import world_module


def run1_gym_world_conversion():
    # a gym world contains white space that is not there in reality. Remove it.
    # A gym world may not be surrounded by walls, add the ability to add them.
    # if walls were added the x,y location of the agent needs to be adjusted accordingly
    gym_world = " o  o  o  o  o \n o  o  x  o  o \n o  o  o  o  o"
    nace_world, xy_loc = world_module.convert_gymnasium_world_to_nace(gym_world)
    assert xy_loc == (3, 2)


if __name__ == "__main__":
    run1_gym_world_conversion()
